# Sample data that is the literal content of /etc/os-release
# for each release, keyed by Docker image (or equivalent)
OS_RELEASE = {
    #######################################################
    # RHEL 8.7
    #######################################################
    "rhel/ubi8:8.7": """\
NAME="Red Hat Enterprise Linux"
VERSION="8.7 (Ootpa)"
ID="rhel"
ID_LIKE="fedora"
VERSION_ID="8.7"
PLATFORM_ID="platform:el8"
PRETTY_NAME="Red Hat Enterprise Linux 8.7 (Ootpa)"
ANSI_COLOR="0;31"
CPE_NAME="cpe:/o:redhat:enterprise_linux:8::baseos"
HOME_URL="https://www.redhat.com/"
DOCUMENTATION_URL="https://access.redhat.com/documentation/red_hat_enterprise_linux/8/"
BUG_REPORT_URL="https://bugzilla.redhat.com/"

REDHAT_BUGZILLA_PRODUCT="Red Hat Enterprise Linux 8"
REDHAT_BUGZILLA_PRODUCT_VERSION=8.7
REDHAT_SUPPORT_PRODUCT="Red Hat Enterprise Linux"
REDHAT_SUPPORT_PRODUCT_VERSION="8.7"
""",
    #######################################################
    # Fedora 37
    #######################################################
    "fedora:37": """\
NAME="Fedora Linux"
VERSION="37 (Container Image)"
ID=fedora
VERSION_ID=37
VERSION_CODENAME=""
PLATFORM_ID="platform:f37"
PRETTY_NAME="Fedora Linux 37 (Container Image)"
ANSI_COLOR="0;38;2;60;110;180"
LOGO=fedora-logo-icon
CPE_NAME="cpe:/o:fedoraproject:fedora:37"
DEFAULT_HOSTNAME="fedora"
HOME_URL="https://fedoraproject.org/"
DOCUMENTATION_URL="https://docs.fedoraproject.org/en-US/fedora/f37/system-administrators-guide/"
SUPPORT_URL="https://ask.fedoraproject.org/"
BUG_REPORT_URL="https://bugzilla.redhat.com/"
REDHAT_BUGZILLA_PRODUCT="Fedora"
REDHAT_BUGZILLA_PRODUCT_VERSION=37
REDHAT_SUPPORT_PRODUCT="Fedora"
REDHAT_SUPPORT_PRODUCT_VERSION=37
SUPPORT_END=2023-11-14
VARIANT="Container Image"
VARIANT_ID=container
""",
    #######################################################
    # AlmaLinux 9.1
    #######################################################
    "almalinux:9.1": """\
NAME="AlmaLinux"
VERSION="9.1 (Lime Lynx)"
ID="almalinux"
ID_LIKE="rhel centos fedora"
VERSION_ID="9.1"
PLATFORM_ID="platform:el9"
PRETTY_NAME="AlmaLinux 9.1 (Lime Lynx)"
ANSI_COLOR="0;34"
LOGO="fedora-logo-icon"
CPE_NAME="cpe:/o:almalinux:almalinux:9::baseos"
HOME_URL="https://almalinux.org/"
DOCUMENTATION_URL="https://wiki.almalinux.org/"
BUG_REPORT_URL="https://bugs.almalinux.org/"

ALMALINUX_MANTISBT_PROJECT="AlmaLinux-9"
ALMALINUX_MANTISBT_PROJECT_VERSION="9.1"
REDHAT_SUPPORT_PRODUCT="AlmaLinux"
REDHAT_SUPPORT_PRODUCT_VERSION="9.1"
""",
    #######################################################
    # Centos 8
    #######################################################
    "centos:8": """\
NAME="CentOS Linux"
VERSION="8"
ID="centos"
ID_LIKE="rhel fedora"
VERSION_ID="8"
PLATFORM_ID="platform:el8"
PRETTY_NAME="CentOS Linux 8"
ANSI_COLOR="0;31"
CPE_NAME="cpe:/o:centos:centos:8"
HOME_URL="https://centos.org/"
BUG_REPORT_URL="https://bugs.centos.org/"
CENTOS_MANTISBT_PROJECT="CentOS-8"
CENTOS_MANTISBT_PROJECT_VERSION="8"
""",
    #######################################################
    # OpenSuSE Leap 15.4
    #######################################################
    "opensuse/leap:15.4": """\
NAME="openSUSE Leap"
VERSION="15.4"
ID="opensuse-leap"
ID_LIKE="suse opensuse"
VERSION_ID="15.4"
PRETTY_NAME="openSUSE Leap 15.4"
ANSI_COLOR="0;32"
CPE_NAME="cpe:/o:opensuse:leap:15.4"
BUG_REPORT_URL="https://bugs.opensuse.org"
HOME_URL="https://www.opensuse.org/"
DOCUMENTATION_URL="https://en.opensuse.org/Portal:Leap"
LOGO="distributor-logo-Leap"
""",
    #######################################################
    # OpenSuse Tumbleweed (20230304)
    #######################################################
    "opensuse/tumbleweed:latest": """\
NAME="openSUSE Tumbleweed"
# VERSION="20230304"
ID="opensuse-tumbleweed"
ID_LIKE="opensuse suse"
VERSION_ID="20230304"
PRETTY_NAME="openSUSE Tumbleweed"
ANSI_COLOR="0;32"
CPE_NAME="cpe:/o:opensuse:tumbleweed:20230304"
BUG_REPORT_URL="https://bugs.opensuse.org"
HOME_URL="https://www.opensuse.org/"
DOCUMENTATION_URL="https://en.opensuse.org/Portal:Tumbleweed"
LOGO="distributor-logo-Tumbleweed"
""",
    #######################################################
    # Debian 11
    #######################################################
    "debian:11": """\
PRETTY_NAME="Debian GNU/Linux 11 (bullseye)"
NAME="Debian GNU/Linux"
VERSION_ID="11"
VERSION="11 (bullseye)"
VERSION_CODENAME=bullseye
ID=debian
HOME_URL="https://www.debian.org/"
SUPPORT_URL="https://www.debian.org/support"
BUG_REPORT_URL="https://bugs.debian.org/"
""",
    #######################################################
    # Ubuntu 22.04 (Jammy Jellyfish)
    #######################################################
    "ubuntu:22.04": """\
PRETTY_NAME="Ubuntu 22.04.2 LTS"
NAME="Ubuntu"
VERSION_ID="22.04"
VERSION="22.04.2 LTS (Jammy Jellyfish)"
VERSION_CODENAME=jammy
ID=ubuntu
ID_LIKE=debian
HOME_URL="https://www.ubuntu.com/"
SUPPORT_URL="https://help.ubuntu.com/"
BUG_REPORT_URL="https://bugs.launchpad.net/ubuntu/"
PRIVACY_POLICY_URL="https://www.ubuntu.com/legal/terms-and-policies/privacy-policy"
UBUNTU_CODENAME=jammy
""",
    #######################################################
    # Pop! OS 22.04
    #######################################################
    "pop:22.04": """\
NAME="Pop!_OS"
VERSION="22.04 LTS"
ID=pop
ID_LIKE="ubuntu debian"
PRETTY_NAME="Pop!_OS 22.04 LTS"
VERSION_ID="22.04"
HOME_URL="https://pop.system76.com"
SUPPORT_URL="https://support.system76.com"
BUG_REPORT_URL="https://github.com/pop-os/pop/issues"
PRIVACY_POLICY_URL="https://system76.com/privacy"
VERSION_CODENAME=jammy
UBUNTU_CODENAME=jammy
LOGO=distributor-logo-pop-os
""",
    #######################################################
    # LinuxMint 19.2
    #######################################################
    "linuxmint:19.2": """\
NAME="Linux Mint"
VERSION="19.2 (Tina)"
ID=linuxmint
ID_LIKE=ubuntu
PRETTY_NAME="Linux Mint 19.2"
VERSION_ID="19.2"
HOME_URL="https://www.linuxmint.com/"
SUPPORT_URL="https://support.ubnutu.com/"
BUG_REPORT_URL="https://linuxmint-troubleshooting-guide.readthedocs.io/en/latest"
PRIVACY_POLICY_URL="https://www.linuxmint.com/"
VERSION_CODENAME=tina
UBUNTU_CODENAME=bionic
""",
    #######################################################
    # Arch Linux (2023-03-03)
    #######################################################
    "archlinux:latest": """\
NAME="Arch Linux"
PRETTY_NAME="Arch Linux"
ID=arch
BUILD_ID=rolling
VERSION_ID=TEMPLATE_VERSION_ID
ANSI_COLOR="38;2;23;147;209"
HOME_URL="https://archlinux.org/"
DOCUMENTATION_URL="https://wiki.archlinux.org/"
SUPPORT_URL="https://bbs.archlinux.org/"
BUG_REPORT_URL="https://bugs.archlinux.org/"
PRIVACY_POLICY_URL="https://terms.archlinux.org/docs/privacy-policy/"
LOGO=archlinux-logo
""",
    #######################################################
    # Manjaro Linux (2023-03-03)
    #######################################################
    "manjaro/base:latest": """\
NAME="Manjaro Linux"
PRETTY_NAME="Manjaro Linux"
ID=manjaro
ID_LIKE=arch
BUILD_ID=rolling
ANSI_COLOR="32;1;24;144;200"
HOME_URL="https://manjaro.org/"
DOCUMENTATION_URL="https://wiki.manjaro.org/"
SUPPORT_URL="https://forum.manjaro.org/"
BUG_REPORT_URL="https://docs.manjaro.org/reporting-bugs/"
PRIVACY_POLICY_URL="https://manjaro.org/privacy-policy/"
LOGO=manjarolinux
""",
}



from __future__ import annotations

from briefcase.config import AppConfig
from briefcase.exceptions import BriefcaseCommandError

from .base import BaseCommand, full_options


class PublishCommand(BaseCommand):
    command = "publish"
    description = "Publish an app to a distribution channel."

    @property
    def publication_channels(self):
        """The list of publication channel backends that are available for this
        format."""
        return ["s3"]

    @property
    def default_publication_channel(self):
        """The default publication channel for this format."""
        return "s3"

    def add_options(self, parser):
        parser.add_argument(
            "-c",
            "--channel",
            choices=self.publication_channels,
            default=self.default_publication_channel,
            help="The channel to publish to",
        )

    def publish_app(self, app: AppConfig, channel: str, **options):
        """Publish an application.

        :param app: The application to publish
        :param channel: The publication channel to use
        """
        self.logger.info(
            f"TODO: Publish {app.app_name} to {channel}"
        )  # pragma: no cover

    def _publish_app(self, app: AppConfig, channel: str, **options) -> dict | None:
        """Internal method to publish a single app. Ensures the app exists, and has been
        packaged before attempting to issue the actual publish command.

        :param app: The application to publish
        :param channel: The publication channel to use
        """
        # TODO: Verify the app has been packaged
        state = None
        self.verify_app(app)

        state = self.publish_app(app, channel=channel, **full_options(state, options))

        return state

    def __call__(
        self,
        channel: str | None = None,
        **options,
    ):
        # Confirm host compatibility, that all required tools are available,
        # and that all app configurations are finalized.
        self.finalize()

        # Check the apps have been built first.
        for app_name, app in self.apps.items():
            binary_file = self.binary_path(app)
            if not binary_file.exists():
                raise BriefcaseCommandError(
                    f"Application {app_name} has not been built. "
                    "Build (and test!) the app before publishing."
                )

        # Then publish them all to the selected channel.
        state = None
        for app_name, app in sorted(self.apps.items()):
            state = self._publish_app(
                app, channel=channel, **full_options(state, options)
            )

        return state



import subprocess

import pytest

from briefcase.exceptions import BriefcaseCommandError


@pytest.mark.parametrize(
    "output, expected_list",
    [
        ("", []),
        ("first\n", ["first"]),
        ("first\nsecond\nthird\n", ["first", "second", "third"]),
        ("first\n\nsecond", ["first", "second"]),
        (
            "first\nINFO    | Storing crashdata in\nsecond\nWARNING | nothing to see\n"
            "third\nERROR   | lot to see here",
            ["first", "second", "third"],
        ),
    ],
)
def test_no_emulators(mock_tools, android_sdk, output, expected_list):
    """The returned list of emulators is properly parsed."""
    mock_tools.subprocess.check_output.return_value = output

    assert android_sdk.emulators() == expected_list


def test_emulator_error(mock_tools, android_sdk):
    """If there is a problem invoking emulator, an error is returned."""
    mock_tools.subprocess.check_output.side_effect = subprocess.CalledProcessError(
        returncode=69, cmd="emulator -list-avd"
    )

    with pytest.raises(
        BriefcaseCommandError,
        match=r"Unable to obtain Android emulator list",
    ):
        android_sdk.emulators()



import pytest

from briefcase.exceptions import UnsupportedHostError
from briefcase.integrations.linuxdeploy import (
    LinuxDeploy,
    LinuxDeployGtkPlugin,
    LinuxDeployQtPlugin,
)

from ...utils import assert_url_resolvable


def test_managed_install(linuxdeploy):
    """All linuxdeploy installs are managed."""
    assert linuxdeploy.managed_install is True


def test_file_path(linuxdeploy, mock_tools):
    """Linuxdeploy resides in the tool path."""
    assert linuxdeploy.file_path == mock_tools.base_path


@pytest.mark.parametrize(
    "host_os, host_arch, is_32bit_python, linuxdeploy_arch",
    [
        ("Linux", "x86_64", False, "x86_64"),
        ("Linux", "x86_64", True, "i386"),
        ("Linux", "i686", True, "i386"),
        ("Linux", "aarch64", True, "armhf"),
        ("Linux", "aarch64", False, "aarch64"),
        ("Linux", "armv7l", True, "armhf"),
        ("Linux", "armv8l", True, "armhf"),
        ("Darwin", "x86_64", False, "x86_64"),
    ],
)
def test_file_name(mock_tools, host_os, host_arch, is_32bit_python, linuxdeploy_arch):
    """Linuxdeploy filename is architecture dependent."""
    mock_tools.host_os = host_os
    mock_tools.host_arch = host_arch
    mock_tools.is_32bit_python = is_32bit_python

    linuxdeploy = LinuxDeploy(mock_tools)

    assert linuxdeploy.file_name == f"linuxdeploy-{linuxdeploy_arch}.AppImage"


def test_file_name_unsupported_arch(mock_tools):
    """LinuxDeploy cannot be verified for an unsupported architecture."""
    mock_tools.host_arch = "IA-64"

    with pytest.raises(
        UnsupportedHostError,
        match="Linux AppImages cannot be built on IA-64.",
    ):
        _ = LinuxDeploy(mock_tools).file_name


@pytest.mark.parametrize(
    "host_os, host_arch, linuxdeploy_arch",
    [
        ("Linux", "x86_64", "x86_64"),
        ("Linux", "i686", "i386"),
        ("Darwin", "x86_64", "x86_64"),
    ],
)
def test_download_url(mock_tools, host_os, host_arch, linuxdeploy_arch):
    """Linuxdeploy download URL is architecture dependent."""
    mock_tools.host_os = host_os
    mock_tools.host_arch = host_arch

    linuxdeploy = LinuxDeploy(mock_tools)

    assert linuxdeploy.download_url == (
        "https://github.com/linuxdeploy/linuxdeploy/"
        f"releases/download/continuous/linuxdeploy-{linuxdeploy_arch}.AppImage"
    )
    assert_url_resolvable(linuxdeploy.download_url)


def test_plugins(linuxdeploy):
    """There are 2 known plugins."""
    assert linuxdeploy.plugins == {
        "gtk": LinuxDeployGtkPlugin,
        "qt": LinuxDeployQtPlugin,
    }



import subprocess
from unittest import mock

import pytest

from briefcase.console import Console, Log, LogLevel
from briefcase.exceptions import UnsupportedHostError
from briefcase.integrations.subprocess import Subprocess
from briefcase.platforms.linux.appimage import LinuxAppImageRunCommand


@pytest.fixture
def run_command(tmp_path):
    command = LinuxAppImageRunCommand(
        logger=Log(),
        console=Console(),
        base_path=tmp_path / "base_path",
        data_path=tmp_path / "briefcase",
    )
    command.tools.home_path = tmp_path / "home"

    # Set the host architecture for test purposes.
    command.tools.host_arch = "x86_64"

    command.tools.subprocess = mock.MagicMock(spec_set=Subprocess)

    command._stream_app_logs = mock.MagicMock()

    return command


@pytest.mark.parametrize("host_os", ["Darwin", "Windows", "WeirdOS"])
def test_unsupported_host_os(run_command, host_os):
    """Error raised for an unsupported OS."""
    run_command.tools.host_os = host_os
    # Mock the existence of a single app
    run_command.apps = {"app": None}

    with pytest.raises(
        UnsupportedHostError,
        match="Linux AppImages can only be executed on Linux.",
    ):
        run_command()


def test_run_gui_app(run_command, first_app_config, tmp_path):
    """A linux GUI App can be started."""
    # Set up the log streamer to return a known stream
    log_popen = mock.MagicMock()
    run_command.tools.subprocess.Popen.return_value = log_popen

    # Run the app
    run_command.run_app(first_app_config, test_mode=False, passthrough=[])

    # The process was started
    run_command.tools.subprocess.Popen.assert_called_with(
        [
            tmp_path
            / "base_path/build/first-app/linux/appimage/First_App-0.0.1-x86_64.AppImage"
        ],
        cwd=tmp_path / "home",
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        bufsize=1,
    )

    # The streamer was started
    run_command._stream_app_logs.assert_called_once_with(
        first_app_config,
        popen=log_popen,
        test_mode=False,
        clean_output=False,
    )


def test_run_gui_app_with_passthrough(run_command, first_app_config, tmp_path):
    """A linux GUI App can be started in debug mode with args."""
    run_command.logger.verbosity = LogLevel.DEBUG

    # Set up the log streamer to return a known stream
    log_popen = mock.MagicMock()
    run_command.tools.subprocess.Popen.return_value = log_popen

    # Run the app with args
    run_command.run_app(
        first_app_config,
        test_mode=False,
        passthrough=["foo", "--bar"],
    )

    # The process was started
    run_command.tools.subprocess.Popen.assert_called_with(
        [
            tmp_path
            / "base_path/build/first-app/linux/appimage/First_App-0.0.1-x86_64.AppImage",
            "foo",
            "--bar",
        ],
        cwd=tmp_path / "home",
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        bufsize=1,
        env={"BRIEFCASE_DEBUG": "1"},
    )

    # The streamer was started
    run_command._stream_app_logs.assert_called_once_with(
        first_app_config,
        popen=log_popen,
        test_mode=False,
        clean_output=False,
    )


def test_run_gui_app_failed(run_command, first_app_config, tmp_path):
    """If there's a problem starting the GUI app, an exception is raised."""
    run_command.tools.subprocess.Popen.side_effect = OSError

    with pytest.raises(OSError):
        run_command.run_app(first_app_config, test_mode=False, passthrough=[])

    # The run command was still invoked
    run_command.tools.subprocess.Popen.assert_called_with(
        [
            tmp_path
            / "base_path/build/first-app/linux/appimage/First_App-0.0.1-x86_64.AppImage"
        ],
        cwd=tmp_path / "home",
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        bufsize=1,
    )

    # No attempt to stream was made
    run_command._stream_app_logs.assert_not_called()


def test_run_console_app(run_command, first_app_config, tmp_path):
    """A linux console App can be started."""
    first_app_config.console_app = True

    # Run the app
    run_command.run_app(first_app_config, test_mode=False, passthrough=[])

    # The process was started
    run_command.tools.subprocess.run.assert_called_with(
        [
            tmp_path
            / "base_path/build/first-app/linux/appimage/First_App-0.0.1-x86_64.AppImage"
        ],
        cwd=tmp_path / "home",
        bufsize=1,
        stream_output=False,
    )

    # No attempt to stream was made
    run_command._stream_app_logs.assert_not_called()


def test_run_console_app_with_passthrough(run_command, first_app_config, tmp_path):
    """A linux console App can be started in debug mode with args."""
    run_command.logger.verbosity = LogLevel.DEBUG

    first_app_config.console_app = True

    # Run the app with args
    run_command.run_app(
        first_app_config,
        test_mode=False,
        passthrough=["foo", "--bar"],
    )

    # The process was started
    run_command.tools.subprocess.run.assert_called_with(
        [
            tmp_path
            / "base_path/build/first-app/linux/appimage/First_App-0.0.1-x86_64.AppImage",
            "foo",
            "--bar",
        ],
        cwd=tmp_path / "home",
        bufsize=1,
        stream_output=False,
        env={"BRIEFCASE_DEBUG": "1"},
    )

    # No attempt to stream was made
    run_command._stream_app_logs.assert_not_called()


def test_run_console_app_failed(run_command, first_app_config, tmp_path):
    """If there's a problem starting the console app, an exception is raised."""
    first_app_config.console_app = True

    run_command.tools.subprocess.run.side_effect = OSError

    with pytest.raises(OSError):
        run_command.run_app(first_app_config, test_mode=False, passthrough=[])

    # The run command was still invoked
    run_command.tools.subprocess.run.assert_called_with(
        [
            tmp_path
            / "base_path/build/first-app/linux/appimage/First_App-0.0.1-x86_64.AppImage"
        ],
        cwd=tmp_path / "home",
        bufsize=1,
        stream_output=False,
    )

    # No attempt to stream was made
    run_command._stream_app_logs.assert_not_called()


@pytest.mark.parametrize("is_console_app", [True, False])
def test_run_app_test_mode(run_command, first_app_config, is_console_app, tmp_path):
    """A linux App can be started in test mode."""
    # Test mode apps are always streamed
    first_app_config.console_app = is_console_app

    # Set up the log streamer to return a known stream
    log_popen = mock.MagicMock()
    run_command.tools.subprocess.Popen.return_value = log_popen

    # Run the app
    run_command.run_app(first_app_config, test_mode=True, passthrough=[])

    # The process was started
    run_command.tools.subprocess.Popen.assert_called_with(
        [
            tmp_path
            / "base_path/build/first-app/linux/appimage/First_App-0.0.1-x86_64.AppImage"
        ],
        cwd=tmp_path / "home",
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        bufsize=1,
        env={"BRIEFCASE_MAIN_MODULE": "tests.first_app"},
    )

    # The streamer was started
    run_command._stream_app_logs.assert_called_once_with(
        first_app_config,
        popen=log_popen,
        test_mode=True,
        clean_output=False,
    )


@pytest.mark.parametrize("is_console_app", [True, False])
def test_run_app_test_mode_with_args(
    run_command,
    first_app_config,
    is_console_app,
    tmp_path,
):
    """A linux App can be started in test mode with args."""
    # Test mode apps are always streamed
    first_app_config.console_app = is_console_app

    # Set up the log streamer to return a known stream
    log_popen = mock.MagicMock()
    run_command.tools.subprocess.Popen.return_value = log_popen

    # Run the app with args
    run_command.run_app(
        first_app_config,
        test_mode=True,
        passthrough=["foo", "--bar"],
    )

    # The process was started
    run_command.tools.subprocess.Popen.assert_called_with(
        [
            tmp_path
            / "base_path/build/first-app/linux/appimage/First_App-0.0.1-x86_64.AppImage",
            "foo",
            "--bar",
        ],
        cwd=tmp_path / "home",
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        bufsize=1,
        env={"BRIEFCASE_MAIN_MODULE": "tests.first_app"},
    )

    # The streamer was started
    run_command._stream_app_logs.assert_called_once_with(
        first_app_config,
        popen=log_popen,
        test_mode=True,
        clean_output=False,
    )



from ...utils import create_file


def test_no_args_package_one_app(package_command, first_app, tmp_path):
    """If there is one app, package that app by default."""
    # Add a single app
    package_command.apps = {
        "first": first_app,
    }

    # Configure no command line options
    options, _ = package_command.parse_options([])

    # Run the run command
    package_command(**options)

    # The right sequence of things will be done
    assert package_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App config has been finalized
        ("finalize-app-config", "first"),
        # App template is verified
        ("verify-app-template", "first"),
        # App tools are verified for app
        ("verify-app-tools", "first"),
        # Package the first app
        (
            "package",
            "first",
            {
                "adhoc_sign": False,
                "identity": None,
            },
        ),
    ]

    # Packaging format has been annotated on the app
    assert first_app.packaging_format == "pkg"

    # The dist folder has been created.
    assert tmp_path / "base_path/dist"


def test_package_one_explicit_app(package_command, first_app, second_app, tmp_path):
    """If one app is named explicitly, package that app."""
    # Add a single app
    package_command.apps = {
        "first": first_app,
        "second": second_app,
    }

    # Configure no command line arguments
    options, _ = package_command.parse_options([])

    # Run the build command on a specific app
    package_command(first_app, **options)

    # The right sequence of things will be done
    assert package_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App config has been finalized
        ("finalize-app-config", "first"),
        # App template is verified
        ("verify-app-template", "first"),
        # App tools are verified for app
        ("verify-app-tools", "first"),
        # Package the first app
        (
            "package",
            "first",
            {
                "adhoc_sign": False,
                "identity": None,
            },
        ),
    ]

    # Packaging format has been annotated on the first app, not the second.
    assert first_app.packaging_format == "pkg"
    assert not hasattr(second_app, "packaging_format")

    # The dist folder has been created.
    assert tmp_path / "base_path/dist"


def test_no_args_package_two_app(package_command, first_app, second_app, tmp_path):
    """If there are multiple apps, publish all of them."""
    # Add two apps
    package_command.apps = {
        "first": first_app,
        "second": second_app,
    }

    # Configure no command line options
    options, _ = package_command.parse_options([])

    # Run the package command
    package_command(**options)

    # The right sequence of things will be done
    assert package_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App configs have been finalized
        ("finalize-app-config", "first"),
        ("finalize-app-config", "second"),
        # App template is verified for first app
        ("verify-app-template", "first"),
        # App tools are verified for first app
        ("verify-app-tools", "first"),
        # Package the first app
        (
            "package",
            "first",
            {
                "adhoc_sign": False,
                "identity": None,
            },
        ),
        # App template is verified for second app
        ("verify-app-template", "second"),
        # App tools are verified for second app
        ("verify-app-tools", "second"),
        # package the second app
        (
            "package",
            "second",
            {
                "adhoc_sign": False,
                "identity": None,
                # state of previous calls have been preserved.
                "package_state": "first",
            },
        ),
    ]

    # Packaging format has been annotated on both apps
    assert first_app.packaging_format == "pkg"
    assert second_app.packaging_format == "pkg"

    # The dist folder has been created.
    assert tmp_path / "base_path/dist"


def test_identity_arg_package_one_app(package_command, first_app, tmp_path):
    """If there is one app,and an --identity argument, package signs the app with the
    specified identity."""
    # Add a single app
    package_command.apps = {
        "first": first_app,
    }

    # Configure an identity option
    options, _ = package_command.parse_options(["--identity", "test"])

    # Run the run command
    package_command(**options)

    # The right sequence of things will be done
    assert package_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App config has been finalized
        ("finalize-app-config", "first"),
        # App template is verified
        ("verify-app-template", "first"),
        # App tools are verified for app
        ("verify-app-tools", "first"),
        # Package the first app
        (
            "package",
            "first",
            {
                "adhoc_sign": False,
                "identity": "test",
            },
        ),
    ]

    # Packaging format has been annotated on the app
    assert first_app.packaging_format == "pkg"

    # The dist folder has been created.
    assert tmp_path / "base_path/dist"


def test_adhoc_sign_package_one_app(package_command, first_app, tmp_path):
    """If there is one app,and an --adhoc argument, package signs the app using ad-hoc
    option."""
    # Add a single app
    package_command.apps = {
        "first": first_app,
    }

    # Configure an ad-hoc signing option
    options, _ = package_command.parse_options(["--adhoc-sign"])

    # Run the run command
    package_command(**options)

    # The right sequence of things will be done
    assert package_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App config has been finalized
        ("finalize-app-config", "first"),
        # App template is verified
        ("verify-app-template", "first"),
        # App tools are verified for app
        ("verify-app-tools", "first"),
        # Package the first app
        (
            "package",
            "first",
            {
                "adhoc_sign": True,
                "identity": None,
            },
        ),
    ]

    # Packaging format has been annotated on the app
    assert first_app.packaging_format == "pkg"

    # The dist folder has been created.
    assert tmp_path / "base_path/dist"


def test_adhoc_sign_args_package_two_app(
    package_command, first_app, second_app, tmp_path
):
    """If there are multiple apps,and an --adhoc argument, package signs all apps using
    ad-hoc identity."""

    package_command.apps = {
        # Add the first app
        "first": first_app,
        # Add the second app
        "second": second_app,
    }

    # Configure adhoc command line options
    options, _ = package_command.parse_options(["--adhoc-sign"])

    # Run the package command
    package_command(**options)

    # The right sequence of things will be done
    assert package_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App configs have been finalized
        ("finalize-app-config", "first"),
        ("finalize-app-config", "second"),
        # App template is verified
        ("verify-app-template", "first"),
        # App tools are verified for first app
        ("verify-app-tools", "first"),
        # Package the first app
        (
            "package",
            "first",
            {
                "adhoc_sign": True,
                "identity": None,
            },
        ),
        # App template is verified
        ("verify-app-template", "second"),
        # App tools are verified for second app
        ("verify-app-tools", "second"),
        # package the second app
        (
            "package",
            "second",
            {
                "adhoc_sign": True,
                "identity": None,
                # state of previous calls have been preserved.
                "package_state": "first",
            },
        ),
    ]

    # Packaging format has been annotated on both apps
    assert first_app.packaging_format == "pkg"
    assert second_app.packaging_format == "pkg"

    # The dist folder has been created.
    assert tmp_path / "base_path/dist"


def test_identity_sign_args_package_two_app(
    package_command, first_app, second_app, tmp_path
):
    """If there are multiple app,and an --identity argument, package signs all the apps
    with the specified identity."""
    # Add a single app
    package_command.apps = {
        "first": first_app,
        "second": second_app,
    }

    # Configure an identity option
    options, _ = package_command.parse_options(["--identity", "test"])

    # Run the run command
    package_command(**options)

    # The right sequence of things will be done
    assert package_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App configs have been finalized
        ("finalize-app-config", "first"),
        ("finalize-app-config", "second"),
        # App template is verified
        ("verify-app-template", "first"),
        # App tools are verified for first app
        ("verify-app-tools", "first"),
        # Package the first app
        (
            "package",
            "first",
            {
                "adhoc_sign": False,
                "identity": "test",
            },
        ),
        # App template is verified
        ("verify-app-template", "second"),
        # App tools are verified for second app
        ("verify-app-tools", "second"),
        # package the second app
        (
            "package",
            "second",
            {
                "adhoc_sign": False,
                "identity": "test",
                # state of previous calls have been preserved.
                "package_state": "first",
            },
        ),
    ]

    # Packaging format has been annotated on both apps
    assert first_app.packaging_format == "pkg"
    assert second_app.packaging_format == "pkg"

    # The dist folder has been created.
    assert tmp_path / "base_path/dist"


def test_package_alternate_format(package_command, first_app, tmp_path):
    """An app can be packaged in an alternate format."""
    # Add a single app
    package_command.apps = {
        "first": first_app,
    }

    # Configure command line options with an alternate format
    options, _ = package_command.parse_options(["--packaging-format", "box"])

    # Run the run command
    package_command(**options)

    # The right sequence of things will be done
    assert package_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App config has been finalized
        ("finalize-app-config", "first"),
        # App template is verified
        ("verify-app-template", "first"),
        # App tools are verified for app
        ("verify-app-tools", "first"),
        # Package the first app
        (
            "package",
            "first",
            {
                "adhoc_sign": False,
                "identity": None,
            },
        ),
    ]

    # Packaging format has been annotated on the app
    assert first_app.packaging_format == "box"

    # The dist folder has been created.
    assert tmp_path / "base_path/dist"


def test_create_before_package(package_command, first_app_config, tmp_path):
    """If the app hasn't been created, package creates the app first."""
    # Add a single app
    package_command.apps = {
        "first": first_app_config,
    }

    # Configure no command line options
    options, _ = package_command.parse_options([])

    # Run the run command
    package_command(**options)

    # The right sequence of things will be done
    assert package_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App config has been finalized
        ("finalize-app-config", "first"),
        # Create and then build the first app
        (
            "create",
            "first",
            {
                "adhoc_sign": False,
                "identity": None,
            },
        ),
        (
            "build",
            "first",
            {
                "adhoc_sign": False,
                "create_state": "first",
                "identity": None,
            },
        ),
        # App template is verified
        ("verify-app-template", "first"),
        # App tools are verified for app
        ("verify-app-tools", "first"),
        # Package the first app
        (
            "package",
            "first",
            {
                "adhoc_sign": False,
                "identity": None,
                "create_state": "first",
                "build_state": "first",
            },
        ),
    ]

    # Packaging format has been annotated on the app
    assert first_app_config.packaging_format == "pkg"

    # The dist folder has been created.
    assert tmp_path / "base_path/dist"


def test_update_package_one_app(package_command, first_app, tmp_path):
    """If there is one app, and a -u argument, package updates the app."""
    # Add a single app
    package_command.apps = {
        "first": first_app,
    }

    # Configure an update option
    options, _ = package_command.parse_options(["-u"])

    # Run the run command
    package_command(**options)

    # The right sequence of things will be done
    assert package_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App config has been finalized
        ("finalize-app-config", "first"),
        # Update (and then build) the first app
        (
            "update",
            "first",
            {
                "adhoc_sign": False,
                "identity": None,
                "update_requirements": True,
                "update_resources": True,
                "update_support": True,
            },
        ),
        (
            "build",
            "first",
            {
                "adhoc_sign": False,
                "identity": None,
                "update_state": "first",
            },
        ),
        # App template is verified
        ("verify-app-template", "first"),
        # App tools are verified for app
        ("verify-app-tools", "first"),
        # Package the first app
        (
            "package",
            "first",
            {
                "adhoc_sign": False,
                "identity": None,
                "update_state": "first",
                "build_state": "first",
            },
        ),
    ]

    # Packaging format has been annotated on the app
    assert first_app.packaging_format == "pkg"

    # The dist folder has been created.
    assert tmp_path / "base_path/dist"


def test_update_package_two_app(package_command, first_app, second_app, tmp_path):
    """If there are multiple apps, update and publish all of them."""
    # Add two apps
    package_command.apps = {
        "first": first_app,
        "second": second_app,
    }

    # Configure an update option
    options, _ = package_command.parse_options(["--update"])

    # Run the package command
    package_command(**options)

    # The right sequence of things will be done
    assert package_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App configs have been finalized
        ("finalize-app-config", "first"),
        ("finalize-app-config", "second"),
        # Update (and then build) the first app
        (
            "update",
            "first",
            {
                "adhoc_sign": False,
                "identity": None,
                "update_requirements": True,
                "update_resources": True,
                "update_support": True,
            },
        ),
        (
            "build",
            "first",
            {
                "adhoc_sign": False,
                "identity": None,
                "update_state": "first",
            },
        ),
        # App template is verified for first app
        ("verify-app-template", "first"),
        # App tools are verified for first app
        ("verify-app-tools", "first"),
        # Package the first app
        (
            "package",
            "first",
            {
                "adhoc_sign": False,
                "identity": None,
                "update_state": "first",
                "build_state": "first",
            },
        ),
        # Update (and then build) the second app
        (
            "update",
            "second",
            {
                "adhoc_sign": False,
                "identity": None,
                "update_requirements": True,
                "update_resources": True,
                "update_support": True,
                # state of previous calls have been preserved.
                "update_state": "first",
                "build_state": "first",
                "package_state": "first",
            },
        ),
        (
            "build",
            "second",
            {
                "adhoc_sign": False,
                "identity": None,
                "update_state": "second",
                # state of previous calls have been preserved.
                "build_state": "first",
                "package_state": "first",
            },
        ),
        # App template is verified for second app
        ("verify-app-template", "second"),
        # App tools are verified for second app
        ("verify-app-tools", "second"),
        # package the second app
        (
            "package",
            "second",
            {
                "adhoc_sign": False,
                "identity": None,
                "update_state": "second",
                "build_state": "second",
                # state of previous calls have been preserved.
                "package_state": "first",
            },
        ),
    ]

    # Packaging format has been annotated on both apps
    assert first_app.packaging_format == "pkg"
    assert second_app.packaging_format == "pkg"

    # The dist folder has been created.
    assert tmp_path / "base_path/dist"


def test_build_before_package(package_command, first_app_unbuilt, tmp_path):
    """If an app hasn't been built, it is built before packaging."""
    # Add a single app
    package_command.apps = {
        "first": first_app_unbuilt,
    }

    # Configure no command line options
    options, _ = package_command.parse_options([])

    # Run the run command
    package_command(**options)

    # The right sequence of things will be done
    assert package_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App config has been finalized
        ("finalize-app-config", "first"),
        # Build the first app
        (
            "build",
            "first",
            {
                "adhoc_sign": False,
                "identity": None,
            },
        ),
        # App template is verified
        ("verify-app-template", "first"),
        # App tools are verified for app
        ("verify-app-tools", "first"),
        # Package the first app
        (
            "package",
            "first",
            {
                "adhoc_sign": False,
                "identity": None,
                "build_state": "first",
            },
        ),
    ]

    # Packaging format has been annotated on the app
    assert first_app_unbuilt.packaging_format == "pkg"

    # The dist folder has been created.
    assert tmp_path / "base_path/dist"


def test_already_packaged(package_command, first_app, tmp_path):
    """If the app has been previously packaged, the old version is deleted."""
    # Add a single app
    package_command.apps = {
        "first": first_app,
    }

    # Mock a historical package artefact.
    artefact_path = tmp_path / "base_path/dist/first-0.0.1.pkg"
    create_file(artefact_path, "Packaged app")

    # Configure no command line options
    options, _ = package_command.parse_options([])

    # Run the run command
    package_command(**options)

    # The right sequence of things will be done
    assert package_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App config has been finalized
        ("finalize-app-config", "first"),
        # App template is verified
        ("verify-app-template", "first"),
        # App tools are verified for app
        ("verify-app-tools", "first"),
        # Package the first app
        (
            "package",
            "first",
            {
                "adhoc_sign": False,
                "identity": None,
            },
        ),
    ]

    # Packaging format has been annotated on the app
    assert first_app.packaging_format == "pkg"

    # The dist folder still exists
    assert tmp_path / "base_path/dist"

    # But the artefact has been deleted.
    # NOTE: This is a testing quirk - because we're mocking the
    # package_app() call, no new artefact is created; the absence
    # of this file shows that the old one has been deleted.
    assert not artefact_path.exists()



from __future__ import annotations

from abc import abstractmethod

from briefcase.config import AppConfig

from .base import BaseCommand, full_options


class OpenCommand(BaseCommand):
    command = "open"
    description = "Open an app in the build tool for the target platform."

    @abstractmethod
    def project_path(self, app: AppConfig):
        """The path in to the project to pass to the shell to open the project."""

    def _open_app(self, app: AppConfig):
        if self.tools.host_os == "Windows":  # pragma: no-cover-if-not-windows
            self.tools.os.startfile(self.project_path(app))
        elif self.tools.host_os == "Darwin":  # pragma: no-cover-if-not-macos
            self.tools.subprocess.Popen(["open", self.project_path(app)])
        else:  # pragma: no-cover-if-not-linux
            self.tools.subprocess.Popen(["xdg-open", self.project_path(app)])

    def open_app(self, app: AppConfig, **options):
        """Open the project for an app.

        :param app: The application to open
        """
        project_path = self.project_path(app)
        if not project_path.exists():
            state = self.create_command(app, **options)
        else:
            state = None

        self.verify_app(app)

        self.logger.info(
            f"Opening {self.project_path(app).relative_to(self.base_path)}...",
            prefix=app.app_name,
        )
        self._open_app(app)

        return state

    def __call__(
        self,
        app: AppConfig | None = None,
        **options,
    ):
        # Confirm host compatibility, that all required tools are available,
        # and that the app configuration is finalized.
        self.finalize(app)

        if app:
            state = self.open_app(app, **options)
        else:
            state = None
            for app_name, app in sorted(self.apps.items()):
                state = self.open_app(app, **full_options(state, options))

        return state



from unittest.mock import MagicMock

import pytest

from briefcase.console import Console, Log
from briefcase.exceptions import BriefcaseConfigError
from briefcase.integrations.flatpak import Flatpak
from briefcase.platforms.linux.flatpak import LinuxFlatpakBuildCommand


@pytest.fixture
def build_command(tmp_path):
    return LinuxFlatpakBuildCommand(
        logger=Log(),
        console=Console(),
        base_path=tmp_path / "base_path",
        data_path=tmp_path / "briefcase",
    )


def test_build(build_command, first_app_config, tmp_path):
    """A flatpak can be built."""
    build_command.tools.flatpak = MagicMock(spec_set=Flatpak)

    first_app_config.flatpak_runtime_repo_url = "https://example.com/flatpak/repo"
    first_app_config.flatpak_runtime_repo_alias = "custom-repo"

    first_app_config.flatpak_runtime = "org.beeware.Platform"
    first_app_config.flatpak_runtime_version = "37.42"
    first_app_config.flatpak_sdk = "org.beeware.SDK"

    build_command.build_app(first_app_config)

    # Repo is verified
    build_command.tools.flatpak.verify_repo.assert_called_once_with(
        repo_alias="custom-repo",
        url="https://example.com/flatpak/repo",
    )

    # Runtimes are verified
    build_command.tools.flatpak.verify_runtime.assert_called_once_with(
        repo_alias="custom-repo",
        runtime="org.beeware.Platform",
        runtime_version="37.42",
        sdk="org.beeware.SDK",
    )

    # The build is invoked
    build_command.tools.flatpak.build.assert_called_once_with(
        bundle_identifier="com.example.first-app",
        app_name="first-app",
        path=tmp_path / "base_path/build/first-app/linux/flatpak",
    )


def test_missing_runtime_config(build_command, first_app_config):
    """The app build errors is a Flatpak runtime is not defined."""
    build_command.tools.flatpak = MagicMock(spec_set=Flatpak)

    with pytest.raises(
        BriefcaseConfigError,
        match="Briefcase configuration error: The App does not specify the Flatpak runtime to use",
    ):
        build_command.build_app(first_app_config)



from unittest.mock import MagicMock

import pytest

import briefcase.integrations.xcode
from briefcase.console import Console, Log
from briefcase.exceptions import NoDistributionArtefact
from briefcase.platforms.iOS.xcode import iOSXcodeCreateCommand


@pytest.fixture
def create_command(tmp_path):
    return iOSXcodeCreateCommand(
        logger=Log(),
        console=Console(),
        base_path=tmp_path / "base_path",
        data_path=tmp_path / "briefcase",
    )


def test_binary_path(create_command, first_app_config, tmp_path):
    binary_path = create_command.binary_path(first_app_config)

    assert binary_path == (
        tmp_path
        / "base_path"
        / "build"
        / "first-app"
        / "ios"
        / "xcode"
        / "build"
        / "Debug-iphonesimulator"
        / "First App.app"
    )


def test_distribution_path(create_command, first_app_config, tmp_path):
    with pytest.raises(
        NoDistributionArtefact,
        match=r"WARNING: No distributable artefact has been generated",
    ):
        create_command.distribution_path(first_app_config)


def test_verify(create_command, monkeypatch):
    """If you're on macOS, you can verify tools."""
    create_command.tools.host_os = "Darwin"

    mock_ensure_xcode_is_installed = MagicMock()
    monkeypatch.setattr(
        briefcase.integrations.xcode.Xcode,
        "ensure_xcode_is_installed",
        mock_ensure_xcode_is_installed,
    )
    mock_ensure_command_line_tools_are_installed = MagicMock()
    monkeypatch.setattr(
        briefcase.integrations.xcode.XcodeCliTools,
        "ensure_command_line_tools_are_installed",
        mock_ensure_command_line_tools_are_installed,
    )
    mock_confirm_xcode_license_accepted = MagicMock()
    monkeypatch.setattr(
        briefcase.integrations.xcode.XcodeCliTools,
        "confirm_xcode_license_accepted",
        mock_confirm_xcode_license_accepted,
    )

    create_command.verify_tools()

    assert create_command.tools.xcode_cli is not None
    mock_ensure_xcode_is_installed.assert_called_once_with(
        tools=create_command.tools,
        min_version=(13, 0, 0),
    )
    mock_ensure_command_line_tools_are_installed.assert_called_once_with(
        tools=create_command.tools
    )
    mock_confirm_xcode_license_accepted.assert_called_once_with(
        tools=create_command.tools
    )



"""Jinja2 extensions."""

import uuid

from jinja2.ext import Extension


class PythonVersionExtension(Extension):
    """Jinja2 extension to convert a full Python version string (3.11.0rc1) into useful
    values."""

    def __init__(self, environment):
        """Initialize the extension with the given environment."""
        super().__init__(environment)

        def py_tag(obj):
            """A Python version tag (3.11)"""
            return ".".join(obj.split(".")[:2])

        def py_libtag(obj):
            """A Python version library tag (311)"""
            return "".join(obj.split(".")[:2])

        environment.filters["py_tag"] = py_tag
        environment.filters["py_libtag"] = py_libtag


class RGBExtension(Extension):
    """Jinja2 extension to convert a hex RGB color to float values."""

    def __init__(self, environment):
        """Initialize the extension with the given environment."""
        super().__init__(environment)

        def float_red(obj):
            try:
                return int(obj.lstrip("#")[:2], 16) / 255.0
            except ValueError:
                return 1.0

        def float_green(obj):
            try:
                return int(obj.lstrip("#")[2:4], 16) / 255.0
            except ValueError:
                return 1.0

        def float_blue(obj):
            try:
                return int(obj.lstrip("#")[4:6], 16) / 255.0
            except ValueError:
                return 1.0

        environment.filters["float_red"] = float_red
        environment.filters["float_green"] = float_green
        environment.filters["float_blue"] = float_blue


class TOMLEscape(Extension):
    """Jinja2 extension to escape strings so TOML don't break."""

    def __init__(self, environment):
        """Initialize the extension with the given environment."""
        super().__init__(environment)

        def escape_toml(obj):
            """Escapes double quotes and backslashes."""
            return obj.replace("\\", "\\\\").replace('"', '\\"')

        def escape_non_ascii(obj):
            """Quotes obj if non ascii characters are present."""
            if obj.isascii():
                return obj
            else:
                return '"' + obj + '"'

        environment.filters["escape_toml"] = escape_toml
        environment.filters["escape_non_ascii"] = escape_non_ascii


class PListExtension(Extension):
    """Jinja2 extension for generating plist values."""

    def __init__(self, environment):
        """Initialize the extension with the given environment."""
        super().__init__(environment)

        def plist_value(obj):
            """Render value in plist format."""
            if isinstance(obj, bool):
                if obj:
                    return "<true/>"
                else:
                    return "<false/>"
            elif isinstance(obj, list):
                children = "\n        ".join(plist_value(value) for value in obj)
                return f"<array>\n        {children}\n    </array>"
            elif isinstance(obj, dict):
                children = "\n        ".join(
                    f"<key>{key}</key>\n        {plist_value(value)}"
                    for key, value in obj.items()
                )
                return f"<dict>\n        {children}\n    </dict>"
            else:
                return f"<string>{obj}</string>"

        environment.filters["plist_value"] = plist_value


class XMLExtension(Extension):
    """Jinja2 extension for generating XML values."""

    def __init__(self, environment):
        """Initialize the extension with the given environment."""
        super().__init__(environment)

        def bool_attr(obj):
            """Render value in XML format appropriate for an attribute."""
            return "true" if obj else "false"

        environment.filters["bool_attr"] = bool_attr


class UUIDExtension(Extension):
    """Extensions for generating UUIDs."""

    def __init__(self, environment):
        """Initialize the extension with the given environment."""
        super().__init__(environment)

        def dns_uuid5(obj):
            """A DNS-based UUID5 object generated from the provided content."""
            return str(uuid.uuid5(uuid.NAMESPACE_DNS, obj))

        environment.filters["dns_uuid5"] = dns_uuid5



import subprocess
from unittest.mock import MagicMock

import pytest

from briefcase.exceptions import BriefcaseCommandError, InvalidDeviceError


def test_install_apk(adb, capsys):
    """Invoking `install_apk()` calls `run()` with the appropriate parameters."""
    # Mock out the run command on an adb instance
    adb.run = MagicMock(return_value="example normal adb output")

    # Invoke install
    adb.install_apk("example.apk")

    # Validate call parameters.
    adb.run.assert_called_once_with("install", "-r", "example.apk")

    # Validate that the normal output of the command was not printed (since there
    # was no error).
    assert "normal adb output" not in capsys.readouterr()


def test_install_failure(adb, capsys):
    """If `install_apk()` fails, an error is raised."""
    # Mock out the run command on an adb instance
    adb.run = MagicMock(
        side_effect=subprocess.CalledProcessError(returncode=2, cmd="install")
    )

    # Invoke install
    with pytest.raises(BriefcaseCommandError):
        adb.install_apk("example.apk")

    # Validate call parameters.
    adb.run.assert_called_once_with("install", "-r", "example.apk")


def test_invalid_device(adb, capsys):
    """Invoking `install_apk()` on an invalid device raises an error."""
    # Mock out the run command on an adb instance
    adb.run = MagicMock(side_effect=InvalidDeviceError("device", "exampleDevice"))

    # Invoke install
    with pytest.raises(InvalidDeviceError):
        adb.install_apk("example.apk")

    # Validate call parameters.
    adb.run.assert_called_once_with("install", "-r", "example.apk")



import pytest

from ....utils import create_file


# Windows' AppConfig requires attribute 'packaging_format'
@pytest.fixture
def first_app_templated(first_app_config, tmp_path):
    app_path = tmp_path / "base_path/build/first-app/windows/app/src"

    # Create the stub binary
    create_file(app_path / "Stub.exe", "Stub binary")

    return first_app_config



from unittest.mock import MagicMock

import pytest

from briefcase.exceptions import UnsupportedHostError
from briefcase.integrations.base import Tool


class DummyTool(Tool):
    """Unmanaged Tool testing class."""

    name = "UnmanagedDummyTool"
    full_name = "Unmanaged Dummy Tool"
    supported_host_os = {"wonky"}

    @classmethod
    def verify_install(cls, tools, **kwargs):
        return f"i'm a {cls.name}"


@pytest.fixture
def unmanaged_tool(mock_tools) -> DummyTool:
    return DummyTool(tools=mock_tools)


@pytest.mark.parametrize(
    "klass, kwargs",
    [
        (DummyTool, {}),
        (DummyTool, {"one": "two", "three": "four"}),
    ],
)
def test_tool_verify(mock_tools, klass, kwargs, monkeypatch):
    """Tool verification checks host OS and tool install."""
    # Wrap verify calls to confirm they were called
    mock_verify_host = MagicMock(wraps=klass.verify_host)
    monkeypatch.setattr(klass, "verify_host", mock_verify_host)
    mock_verify_install = MagicMock(wraps=klass.verify_install)
    monkeypatch.setattr(klass, "verify_install", mock_verify_install)

    # Mock the dummy tool's supported OS
    mock_tools.host_os = "wonky"

    tool = klass.verify(tools=mock_tools, **kwargs)

    mock_verify_host.assert_called_once_with(tools=mock_tools)
    mock_verify_install.assert_called_once_with(tools=mock_tools, app=None, **kwargs)
    assert tool == f"i'm a {klass.name}"


@pytest.mark.parametrize(
    "klass, kwargs",
    [
        (DummyTool, {}),
        (DummyTool, {"one": "two", "three": "four"}),
    ],
)
def test_tool_verify_with_app(mock_tools, first_app_config, klass, kwargs, monkeypatch):
    """App-bound Tool verification checks host OS and tool install."""
    # Wrap verify calls to confirm they were called
    mock_verify_host = MagicMock(wraps=klass.verify_host)
    monkeypatch.setattr(klass, "verify_host", mock_verify_host)
    mock_verify_install = MagicMock(wraps=klass.verify_install)
    monkeypatch.setattr(klass, "verify_install", mock_verify_install)

    # Mock the dummy tool's supported OS
    mock_tools.host_os = "wonky"

    tool = klass.verify(tools=mock_tools, app=first_app_config, **kwargs)

    mock_verify_host.assert_called_once_with(tools=mock_tools)
    mock_verify_install.assert_called_once_with(
        tools=mock_tools,
        app=first_app_config,
        **kwargs,
    )
    assert tool == "i'm a UnmanagedDummyTool"


def test_tool_unsupported_host_os(mock_tools):
    """Tool verification fails for unsupported Host OS."""
    mock_tools.host_os = "not wonky"

    with pytest.raises(
        UnsupportedHostError,
        match="UnmanagedDummyTool is not supported on not wonky",
    ):
        DummyTool.verify(tools=mock_tools)



from unittest import mock

import pytest

from ...utils import create_file


@pytest.fixture
def dummy_app_name():
    return "app_name"


@pytest.fixture
def project_dir_with_files(tmp_path, dummy_app_name, test_source_dir):
    # Setup dummy files created by calling new_app
    out_dir = tmp_path / dummy_app_name
    out_dir.mkdir()

    pyproject_content = """
[tool.briefcase]
placeholder = "abc"
"""
    create_file(out_dir / "pyproject.toml", pyproject_content)
    create_file(out_dir / "CHANGELOG", "CHANGELOG")
    create_file(out_dir / "LICENSE", "LICENSE")
    create_file(out_dir / test_source_dir / f"{dummy_app_name}.py", "test entry script")
    create_file(out_dir / test_source_dir / "test_dummy.py", "dummy tests")
    create_file(out_dir / f"src/{dummy_app_name}/{dummy_app_name}.py", "entry point")
    return out_dir


@pytest.mark.parametrize("test_source_dir", ["test", "tests", "othertest"])
def test_empty_test_source_dir(
    convert_command,
    project_dir_with_files,
    dummy_app_name,
    test_source_dir,
):
    """The full tests dir is copied if no test_source_dir is given."""
    convert_command.migrate_necessary_files(
        project_dir_with_files,
        test_source_dir,
        dummy_app_name,
    )

    dummy_tests = convert_command.base_path / test_source_dir / "test_dummy.py"
    assert dummy_tests.is_file()
    assert dummy_tests.read_text(encoding="utf-8") == "dummy tests"

    test_entry_script = (
        convert_command.base_path / test_source_dir / f"{dummy_app_name}.py"
    )
    assert test_entry_script.is_file()
    assert test_entry_script.read_text(encoding="utf-8") == "test entry script"


@pytest.mark.parametrize("test_source_dir", ["test", "tests", "othertest"])
def test_provided_test_source_dir(
    convert_command,
    project_dir_with_files,
    dummy_app_name,
    test_source_dir,
):
    """Only the test_entry_script is copied if test_source_dir is given."""
    full_test_path = convert_command.base_path / test_source_dir
    full_test_path.mkdir(parents=True)
    convert_command.migrate_necessary_files(
        project_dir_with_files,
        test_source_dir,
        dummy_app_name,
    )

    dummy_tests = full_test_path / "test_dummy.py"
    assert not dummy_tests.is_file()

    test_entry_script = full_test_path / f"{dummy_app_name}.py"
    assert test_entry_script.is_file()
    assert test_entry_script.read_text(encoding="utf-8") == "test entry script"


@pytest.mark.parametrize("test_source_dir", ["tests"])
def test_warning_without_license_file(
    convert_command,
    project_dir_with_files,
    dummy_app_name,
    test_source_dir,
):
    """A single warning is raised if changelog file is present but not license file."""
    convert_command.logger.warning = mock.MagicMock()

    create_file(convert_command.base_path / "CHANGELOG", "")
    convert_command.migrate_necessary_files(
        project_dir_with_files,
        test_source_dir,
        dummy_app_name,
    )

    convert_command.logger.warning.assert_called_once_with(
        f"\nLicense file not found in '{convert_command.base_path}'. "
        "Briefcase will create a template 'LICENSE' file."
    )


@pytest.mark.parametrize("test_source_dir", ["tests"])
def test_file_is_copied_if_no_license_file_specified(
    convert_command,
    project_dir_with_files,
    dummy_app_name,
    test_source_dir,
):
    """A license file is copied if no license file is specified in pyproject.toml."""
    create_file(convert_command.base_path / "CHANGELOG", "")
    assert not (convert_command.base_path / "LICENSE").exists()
    convert_command.migrate_necessary_files(
        project_dir_with_files,
        test_source_dir,
        dummy_app_name,
    )
    assert (convert_command.base_path / "LICENSE").exists()


@pytest.mark.parametrize("test_source_dir", ["tests"])
def test_pep621_specified_license_filename(
    convert_command,
    project_dir_with_files,
    dummy_app_name,
    test_source_dir,
):
    """No license file is copied if a license file is specified in pyproject.toml."""
    convert_command.logger.warning = mock.MagicMock()
    license_name = "LICENSE.txt"
    create_file(convert_command.base_path / license_name, "")
    create_file(
        convert_command.base_path / "pyproject.toml",
        f'[project]\nlicense = {{ file = "{license_name}" }}',
    )
    create_file(convert_command.base_path / "CHANGELOG", "")
    convert_command.migrate_necessary_files(
        project_dir_with_files,
        test_source_dir,
        dummy_app_name,
    )
    assert not (convert_command.base_path / "LICENSE").exists()


@pytest.mark.parametrize("test_source_dir", ["tests"])
def test_pep621_specified_license_text(
    convert_command,
    project_dir_with_files,
    dummy_app_name,
    test_source_dir,
):
    """A license file is copied if the license is specified as text and no LICENSE file
    exists."""
    convert_command.logger.warning = mock.MagicMock()
    create_file(
        convert_command.base_path / "pyproject.toml",
        '[project]\nlicense = { text = "New BSD" }',
    )
    create_file(convert_command.base_path / "CHANGELOG", "")
    convert_command.migrate_necessary_files(
        project_dir_with_files,
        test_source_dir,
        dummy_app_name,
    )
    assert (convert_command.base_path / "LICENSE").exists()

    convert_command.logger.warning.assert_called_once_with(
        f"\nLicense file not found in '{convert_command.base_path}'. "
        "Briefcase will create a template 'LICENSE' file."
    )


@pytest.mark.parametrize("test_source_dir", ["tests"])
def test_warning_without_changelog_file(
    convert_command,
    project_dir_with_files,
    dummy_app_name,
    test_source_dir,
):
    """A single warning is raised if license file is present but not changelog file."""
    convert_command.logger.warning = mock.MagicMock()

    create_file(convert_command.base_path / "LICENSE", "")
    convert_command.migrate_necessary_files(
        project_dir_with_files,
        test_source_dir,
        dummy_app_name,
    )

    convert_command.logger.warning.assert_called_once_with(
        f"\nChangelog file not found in '{convert_command.base_path}'. You should either "
        f"create a new '{convert_command.base_path / 'CHANGELOG'}' file, or rename an "
        "already existing changelog file to 'CHANGELOG'."
    )


@pytest.mark.parametrize("test_source_dir", ["tests"])
def test_no_warning_with_license_and_changelog_file(
    convert_command,
    project_dir_with_files,
    dummy_app_name,
    test_source_dir,
):
    """No warning is raised if both license file and changelog file is present."""
    convert_command.logger.warning = mock.MagicMock()

    create_file(
        convert_command.base_path / "pyproject.toml",
        '[project]\nlicense = { file = "LICENSE.txt" }',
    )
    create_file(convert_command.base_path / "LICENSE.txt", "")
    create_file(convert_command.base_path / "CHANGELOG", "")
    convert_command.migrate_necessary_files(
        project_dir_with_files,
        test_source_dir,
        dummy_app_name,
    )

    convert_command.logger.warning.assert_not_called()


@pytest.mark.parametrize("test_source_dir", ["tests"])
def test_two_warnings_without_license_and_changelog_file(
    convert_command,
    project_dir_with_files,
    dummy_app_name,
    test_source_dir,
):
    """Two warnings are raised if both license file and changelog file are missing."""
    convert_command.logger.warning = mock.MagicMock()

    convert_command.migrate_necessary_files(
        project_dir_with_files,
        test_source_dir,
        dummy_app_name,
    )
    license_warning = (
        f"\nLicense file not found in '{convert_command.base_path}'. "
        "Briefcase will create a template 'LICENSE' file."
    )
    changelog_warning = (
        f"\nChangelog file not found in '{convert_command.base_path}'. You should either "
        f"create a new '{convert_command.base_path / 'CHANGELOG'}' file, or rename an "
        "already existing changelog file to 'CHANGELOG'."
    )
    assert convert_command.logger.warning.mock_calls == [
        mock.call(license_warning),
        mock.call(changelog_warning),
    ]



import subprocess
from unittest import mock

import pytest

from briefcase.console import LogLevel


@pytest.mark.parametrize("tool_debug_mode", (True, False))
def test_run(flatpak, tool_debug_mode):
    """A Flatpak project can be executed."""
    # Enable verbose tool logging
    if tool_debug_mode:
        flatpak.tools.logger.verbosity = LogLevel.DEEP_DEBUG

    # Set up the log streamer to return a known stream
    log_popen = mock.MagicMock()
    flatpak.tools.subprocess.Popen.return_value = log_popen

    # Call run()
    result = flatpak.run(bundle_identifier="com.example.my-app")

    # The expected call was made
    flatpak.tools.subprocess.Popen.assert_called_once_with(
        [
            "flatpak",
            "run",
        ]
        + (["--verbose"] if tool_debug_mode else [])
        + ["com.example.my-app"],
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        bufsize=1,
        **({"env": {"BRIEFCASE_DEBUG": "1"}} if tool_debug_mode else {}),
    )

    # The popen object was returned.
    assert result == log_popen


@pytest.mark.parametrize("tool_debug_mode", (True, False))
def test_run_with_args(flatpak, tool_debug_mode):
    """A Flatpak project can be executed with additional arguments."""
    # Enable verbose tool logging
    if tool_debug_mode:
        flatpak.tools.logger.verbosity = LogLevel.DEEP_DEBUG

    # Set up the log streamer to return a known stream
    log_popen = mock.MagicMock()
    flatpak.tools.subprocess.Popen.return_value = log_popen

    # Call run()
    result = flatpak.run(
        bundle_identifier="com.example.my-app",
        args=["foo", "bar"],
    )

    # The expected call was made
    flatpak.tools.subprocess.Popen.assert_called_once_with(
        [
            "flatpak",
            "run",
        ]
        + (["--verbose"] if tool_debug_mode else [])
        + ["com.example.my-app"]
        + ["foo", "bar"],
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        bufsize=1,
        **({"env": {"BRIEFCASE_DEBUG": "1"}} if tool_debug_mode else {}),
    )

    # The popen object was returned.
    assert result == log_popen


@pytest.mark.parametrize("tool_debug_mode", (True, False))
def test_run_non_streaming(flatpak, tool_debug_mode):
    """A Flatpak project can be executed in non-streaming mode."""
    # Enable verbose tool logging
    if tool_debug_mode:
        flatpak.tools.logger.verbosity = LogLevel.DEEP_DEBUG

    # Call run()
    flatpak.run(
        bundle_identifier="com.example.my-app",
        args=["foo", "bar"],
        stream_output=False,
    )

    # The expected call was made
    flatpak.tools.subprocess.run.assert_called_once_with(
        [
            "flatpak",
            "run",
        ]
        + (["--verbose"] if tool_debug_mode else [])
        + ["com.example.my-app"]
        + ["foo", "bar"],
        bufsize=1,
        stream_output=False,
        **({"env": {"BRIEFCASE_DEBUG": "1"}} if tool_debug_mode else {}),
    )


@pytest.mark.parametrize("tool_debug_mode", (True, False))
def test_main_module_override(flatpak, tool_debug_mode):
    """The main module can be overridden."""
    # Enable verbose tool logging
    if tool_debug_mode:
        flatpak.tools.logger.verbosity = LogLevel.DEEP_DEBUG

    # Set up the log streamer to return a known stream
    log_popen = mock.MagicMock()
    flatpak.tools.subprocess.Popen.return_value = log_popen

    # Call run()
    result = flatpak.run(
        bundle_identifier="com.example.my-app",
        main_module="org.beeware.test-case",
    )

    # The expected call was made
    flatpak.tools.subprocess.Popen.assert_called_once_with(
        [
            "flatpak",
            "run",
        ]
        + (["--verbose"] if tool_debug_mode else [])
        + [
            "com.example.my-app",
        ],
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        bufsize=1,
        env=(
            {
                "BRIEFCASE_MAIN_MODULE": "org.beeware.test-case",
                "BRIEFCASE_DEBUG": "1",
            }
            if tool_debug_mode
            else {
                "BRIEFCASE_MAIN_MODULE": "org.beeware.test-case",
            }
        ),
    )

    # The popen object was returned.
    assert result == log_popen



def test_update_app(update_command, first_app, tmp_path):
    """If the app already exists, it will be updated."""
    update_command.update_app(
        update_command.apps["first"],
        update_requirements=False,
        update_resources=False,
        update_support=False,
        update_stub=False,
        test_mode=False,
    )

    # The right sequence of things will be done
    assert update_command.actions == [
        ("verify-app-template", "first"),
        ("verify-app-tools", "first"),
        ("code", "first", False),
        ("cleanup", "first"),
    ]

    # App content and resources have been updated
    assert (tmp_path / "base_path/build/first/tester/dummy/code.py").exists()
    # requirements and resources haven't been updated
    assert not (tmp_path / "base_path/build/first/tester/dummy/requirements").exists()
    assert not (tmp_path / "base_path/build/first/tester/dummy/resources").exists()
    # Support has not been updated
    assert not (tmp_path / "base_path/build/first/tester/dummy/support").exists()
    # Stub hasn't been updated
    assert not (tmp_path / "base_path/build/first/tester/dummy/stub.exe").exists()
    # ... and the app still exists
    assert (tmp_path / "base_path/build/first/tester/dummy/first.bundle").exists()


def test_update_non_existing_app(update_command, tmp_path):
    """If the app hasn't been generated yet, it won't be created."""

    update_command.update_app(
        update_command.apps["first"],
        update_requirements=False,
        update_resources=False,
        update_support=False,
        update_stub=False,
        test_mode=False,
    )

    # No app creation actions will be performed
    assert update_command.actions == []

    # App content has been not updated
    assert not (tmp_path / "base_path/build/first/tester/dummy/requirements").exists()
    assert not (tmp_path / "base_path/build/first/tester/dummy/code.py").exists()


def test_update_app_with_requirements(update_command, first_app, tmp_path):
    """If the user requests a dependency update, they are updated."""
    update_command.update_app(
        update_command.apps["first"],
        update_requirements=True,
        update_resources=False,
        update_support=False,
        update_stub=False,
        test_mode=False,
    )

    # The right sequence of things will be done
    assert update_command.actions == [
        ("verify-app-template", "first"),
        ("verify-app-tools", "first"),
        ("code", "first", False),
        ("requirements", "first", False),
        ("cleanup", "first"),
    ]

    # App content has been updated
    assert (tmp_path / "base_path/build/first/tester/dummy/requirements").exists()
    assert (tmp_path / "base_path/build/first/tester/dummy/code.py").exists()
    # Extras haven't been updated
    assert not (tmp_path / "base_path/build/first/tester/dummy/resources").exists()
    # Support has not been updated
    assert not (tmp_path / "base_path/build/first/tester/dummy/support").exists()
    # Stub hasn't been updated
    assert not (tmp_path / "base_path/build/first/tester/dummy/stub.exe").exists()
    # ... and the app still exists
    assert (tmp_path / "base_path/build/first/tester/dummy/first.bundle").exists()


def test_update_app_with_resources(update_command, first_app, tmp_path):
    """If the user requests a resources update, they are updated."""
    update_command.update_app(
        update_command.apps["first"],
        update_requirements=False,
        update_resources=True,
        update_support=False,
        update_stub=False,
        test_mode=False,
    )

    # The right sequence of things will be done
    assert update_command.actions == [
        ("verify-app-template", "first"),
        ("verify-app-tools", "first"),
        ("code", "first", False),
        ("resources", "first"),
        ("cleanup", "first"),
    ]

    # App content and resources have been updated
    assert (tmp_path / "base_path/build/first/tester/dummy/code.py").exists()
    assert (tmp_path / "base_path/build/first/tester/dummy/resources").exists()
    # requirements haven't been updated
    assert not (tmp_path / "base_path/build/first/tester/dummy/requirements").exists()
    # Support has not been updated
    assert not (tmp_path / "base_path/build/first/tester/dummy/support").exists()
    # Stub hasn't been updated
    assert not (tmp_path / "base_path/build/first/tester/dummy/stub.exe").exists()
    # ... and the app still exists
    assert (tmp_path / "base_path/build/first/tester/dummy/first.bundle").exists()


def test_update_app_with_support_package(update_command, first_app, tmp_path):
    """If the user requests an app support package update, they are updated."""
    update_command.update_app(
        update_command.apps["first"],
        update_requirements=False,
        update_resources=False,
        update_support=True,
        update_stub=False,
        test_mode=False,
    )

    # The right sequence of things will be done
    assert update_command.actions == [
        ("verify-app-template", "first"),
        ("verify-app-tools", "first"),
        ("code", "first", False),
        ("cleanup-support", "first"),
        ("support", "first"),
        ("cleanup", "first"),
    ]

    # App content and support have been updated
    assert (tmp_path / "base_path/build/first/tester/dummy/code.py").exists()
    # requirements and resources haven't been updated
    assert not (tmp_path / "base_path/build/first/tester/dummy/resources").exists()
    assert not (tmp_path / "base_path/build/first/tester/dummy/requirements").exists()
    # Support has been updated
    assert (tmp_path / "base_path/build/first/tester/dummy/support").exists()
    # Stub hasn't been updated
    assert not (tmp_path / "base_path/build/first/tester/dummy/stub.exe").exists()
    # ... and the app still exists
    assert (tmp_path / "base_path/build/first/tester/dummy/first.bundle").exists()


def test_update_app_with_stub(update_command, first_app, tmp_path):
    """If the user requests an app stub update, it is are updated."""
    # Add an entry to the path index indicating a stub is required
    update_command._briefcase_toml[update_command.apps["first"]] = {
        "paths": {"stub_binary_revision": "b1"}
    }

    update_command.update_app(
        update_command.apps["first"],
        update_requirements=False,
        update_resources=False,
        update_support=False,
        update_stub=True,
        test_mode=False,
    )

    # The right sequence of things will be done
    assert update_command.actions == [
        ("verify-app-template", "first"),
        ("verify-app-tools", "first"),
        ("code", "first", False),
        ("cleanup-stub", "first"),
        ("stub", "first"),
        ("cleanup", "first"),
    ]

    # App content has been updated
    assert (tmp_path / "base_path/build/first/tester/dummy/code.py").exists()
    # requirements and resources haven't been updated
    assert not (tmp_path / "base_path/build/first/tester/dummy/resources").exists()
    assert not (tmp_path / "base_path/build/first/tester/dummy/requirements").exists()
    # Support has not been updated
    assert not (tmp_path / "base_path/build/first/tester/dummy/support").exists()
    # Stub has been updated
    assert (tmp_path / "base_path/build/first/tester/dummy/stub.exe").exists()
    # ... and the app still exists
    assert (tmp_path / "base_path/build/first/tester/dummy/first.bundle").exists()


def test_update_app_stub_without_stub(update_command, first_app, tmp_path):
    """If the user requests an app stub update on an app that doesn't have a stub, it's
    a no-op."""
    update_command.update_app(
        update_command.apps["first"],
        update_requirements=False,
        update_resources=False,
        update_support=False,
        update_stub=True,
        test_mode=False,
    )

    # The right sequence of things will be done
    assert update_command.actions == [
        ("verify-app-template", "first"),
        ("verify-app-tools", "first"),
        ("code", "first", False),
        ("cleanup", "first"),
    ]

    # App content has been updated
    assert (tmp_path / "base_path/build/first/tester/dummy/code.py").exists()
    # requirements and resources haven't been updated
    assert not (tmp_path / "base_path/build/first/tester/dummy/resources").exists()
    assert not (tmp_path / "base_path/build/first/tester/dummy/requirements").exists()
    # Support has not been updated
    assert not (tmp_path / "base_path/build/first/tester/dummy/support").exists()
    # Stub hasn't been updated
    assert not (tmp_path / "base_path/build/first/tester/dummy/stub.exe").exists()
    # ... and the app still exists
    assert (tmp_path / "base_path/build/first/tester/dummy/first.bundle").exists()


def test_update_app_test_mode(update_command, first_app, tmp_path):
    """Update app in test mode."""
    # Pass in the defaults for the update flags
    update_command.update_app(
        update_command.apps["first"],
        test_mode=True,
        update_requirements=False,
        update_resources=False,
        update_support=False,
        update_stub=False,
    )

    # The right sequence of things will be done
    assert update_command.actions == [
        ("verify-app-template", "first"),
        ("verify-app-tools", "first"),
        ("code", "first", True),
        ("cleanup", "first"),
    ]

    # App code has been updated
    assert (tmp_path / "base_path/build/first/tester/dummy/code.py").exists()
    # App requirements and resources have not been updated
    assert not (tmp_path / "base_path/build/first/tester/dummy/requirements").exists()
    assert not (tmp_path / "base_path/build/first/tester/dummy/resources").exists()
    # Support has not been updated
    assert not (tmp_path / "base_path/build/first/tester/dummy/support").exists()
    # Stub hasn't been updated
    assert not (tmp_path / "base_path/build/first/tester/dummy/stub.exe").exists()
    # ... and the app still exists
    assert (tmp_path / "base_path/build/first/tester/dummy/first.bundle").exists()


def test_update_app_test_mode_requirements(update_command, first_app, tmp_path):
    """Update app in test mode, but with requirements."""
    # Pass in the defaults for the update flags
    update_command.update_app(
        update_command.apps["first"],
        test_mode=True,
        update_requirements=True,
        update_resources=False,
        update_support=False,
        update_stub=False,
    )

    # The right sequence of things will be done
    assert update_command.actions == [
        ("verify-app-template", "first"),
        ("verify-app-tools", "first"),
        ("code", "first", True),
        ("requirements", "first", True),
        ("cleanup", "first"),
    ]

    # App content and requirements have been updated
    assert (tmp_path / "base_path/build/first/tester/dummy/code.py").exists()
    assert (tmp_path / "base_path/build/first/tester/dummy/requirements").exists()
    # App resources have not been updated
    assert not (tmp_path / "base_path/build/first/tester/dummy/resources").exists()
    # Support has not been updated
    assert not (tmp_path / "base_path/build/first/tester/dummy/support").exists()
    # Stub hasn't been updated
    assert not (tmp_path / "base_path/build/first/tester/dummy/stub.exe").exists()
    # ... and the app still exists
    assert (tmp_path / "base_path/build/first/tester/dummy/first.bundle").exists()


def test_update_app_test_mode_resources(update_command, first_app, tmp_path):
    """Update app in test mode, but with resources."""
    # Pass in the defaults for the update flags
    update_command.update_app(
        update_command.apps["first"],
        test_mode=True,
        update_requirements=False,
        update_resources=True,
        update_support=False,
        update_stub=False,
    )

    # The right sequence of things will be done
    assert update_command.actions == [
        ("verify-app-template", "first"),
        ("verify-app-tools", "first"),
        ("code", "first", True),
        ("resources", "first"),
        ("cleanup", "first"),
    ]

    # App content and resources have been updated
    assert (tmp_path / "base_path/build/first/tester/dummy/code.py").exists()
    assert (tmp_path / "base_path/build/first/tester/dummy/resources").exists()
    # App requirements have not been updated
    assert not (tmp_path / "base_path/build/first/tester/dummy/requirements").exists()
    # Support has not been updated
    assert not (tmp_path / "base_path/build/first/tester/dummy/support").exists()
    # Stub hasn't been updated
    assert not (tmp_path / "base_path/build/first/tester/dummy/stub.exe").exists()
    # ... and the app still exists
    assert (tmp_path / "base_path/build/first/tester/dummy/first.bundle").exists()



import pytest

from briefcase.console import Console, Log
from briefcase.platforms.web.static import StaticWebCreateCommand


@pytest.fixture
def create_command(tmp_path):
    return StaticWebCreateCommand(
        logger=Log(),
        console=Console(),
        base_path=tmp_path / "base_path",
        data_path=tmp_path / "briefcase",
    )


def test_project_path(create_command, first_app_config, tmp_path):
    project_path = create_command.project_path(first_app_config)

    assert project_path == (tmp_path / "base_path/build/first-app/web/static/www")


def test_binary_path(create_command, first_app_config, tmp_path):
    binary_path = create_command.binary_path(first_app_config)

    assert binary_path == (
        tmp_path
        / "base_path"
        / "build"
        / "first-app"
        / "web"
        / "static"
        / "www"
        / "index.html"
    )


def test_wheel_path(create_command, first_app_config, tmp_path):
    wheel_path = create_command.wheel_path(first_app_config)

    assert wheel_path == (
        tmp_path
        / "base_path"
        / "build"
        / "first-app"
        / "web"
        / "static"
        / "www"
        / "static"
        / "wheels"
    )


def test_distribution_path(create_command, first_app_config, tmp_path):
    distribution_path = create_command.distribution_path(first_app_config)

    assert distribution_path == (tmp_path / "base_path/dist/First App-0.0.1.web.zip")



from .conftest import DummyCommand


def test_briefcase_required_python_version(base_command):
    assert base_command.briefcase_required_python_version == (3, 8)


def test_bundle_path(base_command, my_app, tmp_path):
    bundle_path = base_command.bundle_path(my_app)

    assert bundle_path == tmp_path / "base_path/build/my-app/tester/dummy"


def test_create_command(base_command):
    # Check for a property of the created command class.
    assert base_command.create_command.description == "Test Create"


def test_update_command(base_command):
    # Check for a property of the created command class.
    assert base_command.update_command.description == "Test Update"


def test_build_command(base_command):
    # Check for a property of the created command class.
    assert base_command.build_command.description == "Test Build"


def test_run_command(base_command):
    # Check for a property of the created command class.
    assert base_command.run_command.description == "Test Run"


def test_package_command(base_command):
    # Check for a property of the created command class.
    assert base_command.package_command.description == "Test Package"


def test_publish_command(base_command):
    # Check for a property of the created command class.
    assert base_command.publish_command.description == "Test Publish"


def test_command_state_transferred(tmp_path):
    """Command state is transferred to created subcommands."""
    command = DummyCommand(base_path=tmp_path)
    command.tools.input.enabled = False

    # Check the enabled state of subcommands
    assert not command.create_command.input.enabled
    assert command.create_command.logger is command.tools.logger
    assert command.create_command.input is command.tools.input
    assert command.create_command.tools is command.tools
    assert command.create_command.is_clone is True

    assert not command.update_command.input.enabled
    assert command.update_command.logger is command.tools.logger
    assert command.update_command.input is command.tools.input
    assert command.update_command.tools is command.tools
    assert command.update_command.is_clone is True

    assert not command.build_command.input.enabled
    assert command.build_command.logger is command.tools.logger
    assert command.build_command.input is command.tools.input
    assert command.build_command.tools is command.tools
    assert command.build_command.is_clone is True

    assert not command.run_command.input.enabled
    assert command.run_command.logger is command.tools.logger
    assert command.run_command.input is command.tools.input
    assert command.run_command.tools is command.tools
    assert command.run_command.is_clone is True

    assert not command.package_command.input.enabled
    assert command.package_command.logger is command.tools.logger
    assert command.package_command.input is command.tools.input
    assert command.package_command.tools is command.tools
    assert command.package_command.is_clone is True

    assert not command.publish_command.input.enabled
    assert command.publish_command.logger is command.tools.logger
    assert command.publish_command.input is command.tools.input
    assert command.publish_command.tools is command.tools
    assert command.publish_command.is_clone is True



from unittest.mock import MagicMock

import pytest

from briefcase.exceptions import (
    MissingToolError,
    NonManagedToolError,
    UnsupportedHostError,
)
from briefcase.integrations.base import ManagedTool


class DummyManagedTool(ManagedTool):
    """Managed Tool testing class."""

    name = "ManagedDummyTool"
    full_name = "Managed Dummy Tool"
    supported_host_os = {"wonky"}

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.actions = []

    @classmethod
    def verify_install(cls, tools, **kwargs):
        return f"i'm a {cls.name}"

    def exists(self) -> bool:
        self.actions.append("exists")
        return True

    def install(self):
        self.actions.append("install")

    def uninstall(self):
        self.actions.append("uninstall")


class NonExistsManagedTool(DummyManagedTool):
    def exists(self) -> bool:
        return False


class NonManagedManagedTool(DummyManagedTool):
    @property
    def managed_install(self) -> bool:
        return False


@pytest.fixture
def managed_tool(mock_tools) -> DummyManagedTool:
    return DummyManagedTool(tools=mock_tools)


@pytest.fixture
def nonexistent_managed_tool(mock_tools) -> NonExistsManagedTool:
    return NonExistsManagedTool(tools=mock_tools)


@pytest.fixture
def nonmanaged_managed_tool(mock_tools) -> NonManagedManagedTool:
    return NonManagedManagedTool(tools=mock_tools)


@pytest.mark.parametrize(
    "klass, kwargs",
    [
        (DummyManagedTool, {}),
        (DummyManagedTool, {"one": "two", "three": "four"}),
    ],
)
def test_tool_verify(mock_tools, klass, kwargs, monkeypatch):
    """Tool verification checks host OS and tool install."""
    # Wrap verify calls to confirm they were called
    mock_verify_host = MagicMock(wraps=klass.verify_host)
    monkeypatch.setattr(klass, "verify_host", mock_verify_host)
    mock_verify_install = MagicMock(wraps=klass.verify_install)
    monkeypatch.setattr(klass, "verify_install", mock_verify_install)

    # Mock the dummy tool's supported OS
    mock_tools.host_os = "wonky"

    tool = klass.verify(tools=mock_tools, **kwargs)

    mock_verify_host.assert_called_once_with(tools=mock_tools)
    mock_verify_install.assert_called_once_with(
        tools=mock_tools, app=None, install=True, **kwargs
    )
    assert tool == f"i'm a {klass.name}"


@pytest.mark.parametrize(
    "klass, kwargs",
    [
        (DummyManagedTool, {}),
        (DummyManagedTool, {"one": "two", "three": "four"}),
    ],
)
def test_tool_verify_with_app(mock_tools, first_app_config, klass, kwargs, monkeypatch):
    """App-bound Tool verification checks host OS and tool install."""
    # Wrap verify calls to confirm they were called
    mock_verify_host = MagicMock(wraps=klass.verify_host)
    monkeypatch.setattr(klass, "verify_host", mock_verify_host)
    mock_verify_install = MagicMock(wraps=klass.verify_install)
    monkeypatch.setattr(klass, "verify_install", mock_verify_install)

    # Mock the dummy tool's supported OS
    mock_tools.host_os = "wonky"

    tool = klass.verify(tools=mock_tools, app=first_app_config, **kwargs)

    mock_verify_host.assert_called_once_with(tools=mock_tools)
    mock_verify_install.assert_called_once_with(
        tools=mock_tools,
        app=first_app_config,
        install=True,
        **kwargs,
    )
    assert tool == "i'm a ManagedDummyTool"


def test_tool_unsupported_host_os(mock_tools):
    """Tool verification fails for unsupported Host OS."""
    mock_tools.host_os = "not wonky"

    with pytest.raises(
        UnsupportedHostError,
        match="ManagedDummyTool is not supported on not wonky",
    ):
        DummyManagedTool.verify(tools=mock_tools)


def test_managed_install_is_true(managed_tool):
    """Tool.managed_install defaults False."""
    assert managed_tool.managed_install is True


def test_managed_upgrade(managed_tool):
    """Order of operations is correct for upgrade."""
    managed_tool.upgrade()

    assert managed_tool.actions == ["exists", "uninstall", "install"]


def test_managed_raises_if_unmanaged(mock_tools, nonmanaged_managed_tool):
    """If a ManagedTool is unmanaged, upgrade raises."""
    with pytest.raises(NonManagedToolError):
        nonmanaged_managed_tool.upgrade()


def test_managed_raises_if_not_exists(mock_tools, nonexistent_managed_tool):
    """If a ManagedTool doesn't exist, upgrade raises."""
    with pytest.raises(MissingToolError):
        nonexistent_managed_tool.upgrade()



from __future__ import annotations

import enum
import re
import subprocess
from pathlib import Path

from briefcase.exceptions import BriefcaseCommandError, CommandOutputParseError
from briefcase.integrations.base import Tool, ToolCache
from briefcase.integrations.subprocess import json_parser


class DeviceState(enum.Enum):
    SHUTDOWN = 0
    BOOTED = 1
    SHUTTING_DOWN = 10
    UNKNOWN = 99


class Xcode(Tool):
    name = "xcode"
    full_name = "Xcode"
    supported_host_os = {"Darwin"}

    @classmethod
    def verify_install(
        cls,
        tools: ToolCache,
        min_version: tuple[int, int] | tuple[int, int, int] | None = None,
        **kwargs,
    ) -> Xcode:
        """Verify that Xcode and the command line developer tools are installed and
        ready for use.

        We need Xcode, *and* the Xcode Command Line Tools. A completely clean
        machine will have neither Xcode *nor* the Command Line Tools. However,
        it's possible to install Xcode and *not* install the command line tools,
        and vice versa.

        We also need to ensure that an adequate version of Xcode is available.

        Then, there is a license that needs to be accepted.

        Lastly, we ensure that the iOS simulator is installed.

        :param tools: ToolCache of available tools
        :param min_version: The minimum allowed version of Xcode, specified as a
            tuple of integers (e.g., (11, 2, 1)). Default: ``None``, meaning there
            is no minimum version.
        """
        # short circuit since already verified and available
        if hasattr(tools, "xcode"):
            return tools.xcode

        cls.ensure_xcode_is_installed(tools=tools, min_version=min_version)
        XcodeCliTools.verify(tools=tools)

        tools.xcode = Xcode(tools=tools)
        return tools.xcode

    @classmethod
    def ensure_xcode_is_installed(
        cls,
        tools: ToolCache,
        min_version: tuple[int, int] | tuple[int, int, int] | None = None,
        xcode_location: str = "/Applications/Xcode.app",
    ):
        """Determine if Xcode is installed; and if so, that it meets minimum version
        requirements.

        Raises an exception if Xcode isn't installed, or if the version of Xcode
        that is installed doesn't meet the minimum requirement.

        :param tools: ToolCache of available tools
        :param min_version: The minimum allowed version of Xcode, specified as a
            tuple of integers (e.g., (11, 2, 1)). Default: ``None``, meaning there
            is no minimum version.
        :param xcode_location: The location where we expect to find an Xcode install.
            Used for testing; defaults to ``/Applications/Xcode.app``.
        """
        # Check for *any* version of Xcode tools. xcode-select returns:
        #  * The path to the currently active Xcode install; or
        #  * error code 2 - No Xcode installation
        try:
            tools.subprocess.check_output(["xcode-select", "-p"])
        except subprocess.CalledProcessError as e:
            raise BriefcaseCommandError(
                """\
Could not find an Xcode installation.

To select an existing Xcode installation, run:

    $ sudo xcode-select --switch path/to/Xcode.app

or install Xcode from the macOS App Store. Once you have installed Xcode,
you can re-run Briefcase.
"""
            ) from e

        try:
            # xcodebuild -version returns the version of Xcode that is currently
            # selected. If the current Xcode is a commandline tools install,
            # returns an error:
            #   xcode-select: error: tool 'xcodebuild' requires Xcode, but active
            #   developer directory '/Library/Developer/CommandLineTools' is a
            #   command line tools instance
            output = tools.subprocess.check_output(["xcodebuild", "-version"])

            if min_version is not None:
                # Look for a line in the output that reads "Xcode X.Y.Z"
                version_lines = [
                    line for line in output.split("\n") if line.startswith("Xcode ")
                ]
                if version_lines:
                    # Split the content after the first space
                    # and split that content on the dots.
                    # Append 0's to fill any gaps caused by
                    # version numbers that don't have a minor version.
                    # At this point, version lines *must* have at least one element,
                    # and each line *must* have a string with at least one space,
                    # so if either array lookup fails, something weird is happening.
                    version = tuple(
                        int(v) for v in version_lines[0].split(" ")[1].split(".")
                    ) + (0, 0)

                    if version < min_version:
                        min_version = ".".join(str(v) for v in min_version)
                        version = ".".join(str(v) for v in version)
                        raise BriefcaseCommandError(
                            f"Xcode {min_version} is required; {version} is installed. Please update Xcode."
                        )
                    else:
                        # Version number is acceptable
                        return

                tools.logger.warning(
                    """
*************************************************************************
** WARNING: Unable to determine the version of Xcode that is installed **
*************************************************************************

    Briefcase will proceed, assuming everything is OK. If you experience
    problems, this is almost certainly the cause of those problems.

    Please report this as a bug at:

      https://github.com/beeware/briefcase/issues/new

    In your report, please including the output from running:

        $ xcodebuild -version

    from the command prompt.

*************************************************************************
"""
                )

        except subprocess.CalledProcessError as e:
            if " is a command line tools instance" in e.output:
                # Commandline tools are currently selected. Look for the existence
                # of the default folder; if that folder doesn't exist, we can't
                # conclude that Xcode *isn't* installed.
                if Path(xcode_location).exists():
                    preamble = """\
Xcode appears to be installed, but the active developer directory is the Xcode
command line tools. To make Xcode the active developer directory, run:

    $ sudo xcode-select --switch /Applications/Xcode.app
"""
                else:
                    preamble = """\
You have the Xcode command line tools installed; however, Briefcase requires
a full Xcode install. Xcode can be downloaded from the macOS App Store at
<https://apps.apple.com/au/app/xcode/id497799835?mt=12>.
"""

                raise BriefcaseCommandError(
                    preamble
                    + """
Or, to use a version of Xcode installed in a non-default location:

    $ sudo xcode-select --switch /path/to/Xcode.app

and then re-run Briefcase.
"""
                ) from e

            else:
                raise BriefcaseCommandError(
                    """\
An Xcode install appears to exist, but Briefcase was unable to
determine the current Xcode version. Running:

    $ xcodebuild -version

should return the current Xcode version, but it raised an error.

You may need to re-install Xcode. Re-run Briefcase once that
installation is complete.
"""
                ) from e


class XcodeCliTools(Tool):
    name = "xcode_cli"
    full_name = "Xcode Command Line Tools"
    supported_host_os = {"Darwin"}

    @classmethod
    def verify_install(cls, tools: ToolCache, **kwargs) -> XcodeCliTools:
        """Verify that command line developer tools are installed and ready for use.

        A completely clean machine will have neither Xcode *nor* the Command Line Tools.
        However, it's possible to install Xcode and *not* install the command line
        tools, and vice versa.

        Lastly, there is a license that needs to be accepted.

        :param tools: ToolCache of available tools
        """
        # short circuit since already verified and available
        if hasattr(tools, "xcode_cli"):
            return tools.xcode_cli

        cls.ensure_command_line_tools_are_installed(tools=tools)
        cls.confirm_xcode_license_accepted(tools=tools)

        tools.xcode_cli = XcodeCliTools(tools=tools)
        return tools.xcode_cli

    @classmethod
    def ensure_command_line_tools_are_installed(cls, tools: ToolCache):
        """Determine if the Xcode command line tools are installed.

        If they are not installed, an exception is raised; in addition, an OS dialog
        will be displayed prompting the user to install Xcode.

        :param tools: ToolCache of available tools
        """
        # We determine if the command line tools are installed by running:
        #
        #   xcode-select --install
        #
        # If that command exits with status 0, it means the tools are *not*
        # installed; but a dialog will be displayed prompting an installation.
        #
        # If it returns a status code of 1, the tools are already installed
        # and outputs the message "command line tools are already installed"
        #
        # Any other status code is a problem.
        try:
            tools.subprocess.check_output(["xcode-select", "--install"])
            raise BriefcaseCommandError(
                """\
The command line developer tools are not installed.

You should be shown a dialog prompting you to install them. Select "Install"
to continue, and re-run Briefcase once that installation is complete.
"""
            )
        except subprocess.CalledProcessError as e:
            if e.returncode != 1:
                tools.logger.warning(
                    """
*************************************************************************
** WARNING: Unable to determine if Xcode is installed                  **
*************************************************************************

    Briefcase will proceed, assuming everything is OK. If you experience
    problems, this is almost certainly the cause of those problems.

    Please report this as a bug at:

       https://github.com/beeware/briefcase/issues/new

    In your report, please including the output from running:

        $ xcode-select --install

    from the command prompt.

*************************************************************************
"""
                )

    @classmethod
    def confirm_xcode_license_accepted(cls, tools: ToolCache):
        """Confirm if the Xcode license has been accepted.

        :param tools: ToolCache of available tools
        """
        # Lastly, check if the Xcode license has been accepted. The command line
        # tools return a status code of 69 (nice...) if the license has not been
        # accepted. In this case, we can prompt the user to accept the license.
        try:
            tools.subprocess.check_output(["/usr/bin/clang", "--version"])
        except subprocess.CalledProcessError as e:
            if e.returncode == 69:
                tools.logger.info(
                    """
Use of Xcode and the iOS developer tools are covered by a license that must be
accepted before you can use those tools.

You can accept these licenses by starting Xcode and clicking "Accept"; or, you
can run this command and accept the license when prompted:

    $ sudo xcodebuild -license

Briefcase will try to run this command now. You will need to enter your
password (Briefcase will not store this password anywhere).
"""
                )
                try:
                    tools.subprocess.run(
                        ["sudo", "xcodebuild", "-license"],
                        check=True,
                        stream_output=False,
                    )
                except subprocess.CalledProcessError as e:
                    # status code 1 - sudo fail
                    # status code 69 - license not accepted.
                    if e.returncode == 1:
                        raise BriefcaseCommandError(
                            """\
Briefcase was unable to run the Xcode licensing tool. This may be because you
did not enter your password correctly, or because your account does not have
administrator privileges on this computer.

You need to accept the Xcode license before Briefcase can package your app.
"""
                        )
                    elif e.returncode == 69:
                        raise BriefcaseCommandError(
                            """\
Xcode license has not been accepted. Briefcase cannot continue.

You need to accept the Xcode license before Briefcase can package your app.
"""
                        )
                    else:
                        tools.logger.warning(
                            """
*************************************************************************
** WARNING: Unable to determine if the Xcode license has been accepted **
*************************************************************************

    Briefcase will proceed, assuming everything is OK. If you experience
    problems, this is almost certainly the cause of those problems.

    Please report this as a bug at:

      https://github.com/beeware/briefcase/issues/new

    In your report, please including the output from running:

        $ sudo xcodebuild -license

    from the command prompt.

*************************************************************************
"""
                        )
            else:
                tools.logger.warning(
                    """
*************************************************************************
** WARNING: Unable to determine if the Xcode license has been accepted **
*************************************************************************

    Briefcase will proceed, assuming everything is OK. If you experience
    problems, this is almost certainly the cause of those problems.

    Please report this as a bug at:

      https://github.com/beeware/briefcase/issues/new

    In your report, please including the output from running:

        $ /usr/bin/clang --version

    from the command prompt.

*************************************************************************
"""
                )


def get_simulators(
    tools: ToolCache,
    os_name: str,
    simulator_location: str = "/Library/Developer/PrivateFrameworks/CoreSimulator.framework/",
) -> dict[str, str]:
    """Obtain the simulators available on this machine.

    The return value is a 2 level dictionary. The outer dictionary is keyed by OS
    version; the inner dictionary for each OS version contains the details of the
    available simulators, keyed by UDID.

    :param tools: ToolCache of available tools
    :param os_name: The OS that we want to simulate. One of `"iOS"`, `"watchOS"`, or
        `"tvOS"`.
    :param simulator_location: The filesystem path where the simulator frameworks are
        installed.
    :returns: A dictionary of available simulators.
    """
    # If the simulator frameworks don't exist, they will be downloaded
    # and installed. This should only occur on first execution.
    if not Path(simulator_location).exists():
        tools.input.prompt(
            f"""
It looks like the {os_name} Simulator is not installed. The {os_name} Simulator
must be installed with administrator privileges.

xcodebuild will prompt you for your admin password so that it can download
and install the simulator.
"""
        )
        tools.input("Press Return to continue: ")

    try:
        simctl_data = tools.subprocess.parse_output(
            json_parser,
            ["xcrun", "simctl", "list", "-j"],
        )

        os_versions = {
            runtime["name"]: runtime["identifier"]
            for runtime in simctl_data["runtimes"]
            if runtime["name"].startswith(f"{os_name} ") and runtime["isAvailable"]
        }

        # For some reason, simctl varies the style of key that is used to
        # identify device versions. The first format is OS identifier (e.g.,
        # 'com.apple.CoreSimulator.SimRuntime.iOS-12-0'). The second is a
        # "human readable" name ('iOS 12.0'). We presume (but can't verify)
        # that any given OS version only exists with a single key.
        # SO - Look for an identifier first; then look for the OS name. If
        # neither exist, return an empty list.
        simulators = {
            version: {
                device["udid"]: device["name"]
                for device in simctl_data["devices"].get(
                    identifier, simctl_data["devices"].get(version, [])
                )
                if device["isAvailable"]
            }
            for version, identifier in os_versions.items()
        }

        # Purge any versions with no devices
        versions_with_no_devices = [
            version for version, devices in simulators.items() if len(devices) == 0
        ]
        for version in versions_with_no_devices:
            simulators.pop(version)

        return simulators

    except CommandOutputParseError as e:
        raise BriefcaseCommandError("Unable to parse output of xcrun simctl") from e
    except subprocess.CalledProcessError as e:
        raise BriefcaseCommandError("Unable to run xcrun simctl.") from e


def get_device_state(tools: ToolCache, udid: str) -> str:
    """Determine the state of an iOS simulator device.

    :param tools: ToolCache of available tools
    :param udid: The UDID of the device to inspect
    :returns: The status of the device, as a DeviceState enum.
    """
    try:
        simctl_data = tools.subprocess.parse_output(
            json_parser,
            ["xcrun", "simctl", "list", "devices", "-j", udid],
        )

        for runtime, devices in simctl_data["devices"].items():
            for device in devices:
                if device["udid"] == udid:
                    return {
                        "Booted": DeviceState.BOOTED,
                        "Shutting Down": DeviceState.SHUTTING_DOWN,
                        "Shutdown": DeviceState.SHUTDOWN,
                    }.get(device["state"], DeviceState.UNKNOWN)

        # If we fall out the bottom of the loop, the UDID didn't match
        # so we raise an error.
        raise BriefcaseCommandError(f"Unable to determine status of device {udid}.")
    except CommandOutputParseError as e:
        raise BriefcaseCommandError("Unable to parse output of xcrun simctl") from e
    except subprocess.CalledProcessError as e:
        raise BriefcaseCommandError("Unable to run xcrun simctl.") from e


# A regex pattern that matches the content returned by `security find-identity`
IDENTITY_RE = re.compile(r"\s*\d+\) ([0-9A-F]{40}) \"(.*)\"")


def get_identities(tools: ToolCache, policy: str | None = None) -> dict[str, str]:
    """Obtain a set of valid identities for the given policy.

    :param tools: ToolCache of available tools
    :param policy: The identity policy to evaluate (e.g., ``codesigning``).
        If None, no policy filtering is applied.
    """
    try:
        output = tools.subprocess.check_output(
            ["security", "find-identity", "-v"] + (["-p", policy] if policy else []),
        )

        return dict(
            IDENTITY_RE.match(line).groups()
            for line in output.split("\n")
            if IDENTITY_RE.match(line)
        )

    except subprocess.CalledProcessError as e:
        raise BriefcaseCommandError("Unable to run security find-identity.") from e



from unittest.mock import MagicMock

import pytest

from briefcase.console import Console, Log
from briefcase.exceptions import BriefcaseConfigError, UnsupportedHostError
from briefcase.integrations.flatpak import Flatpak
from briefcase.platforms.linux.flatpak import LinuxFlatpakCreateCommand


@pytest.fixture
def create_command(tmp_path):
    return LinuxFlatpakCreateCommand(
        logger=Log(),
        console=Console(),
        base_path=tmp_path / "base_path",
        data_path=tmp_path / "briefcase",
    )


@pytest.mark.parametrize("host_os", ["Darwin", "Windows", "WeirdOS"])
def test_unsupported_host_os(create_command, host_os):
    """Error raised for an unsupported OS."""
    create_command.tools.host_os = host_os

    with pytest.raises(
        UnsupportedHostError,
        match="Flatpaks can only be built on Linux.",
    ):
        create_command()


def test_output_format_template_context(create_command, first_app_config):
    """The template context is provided flatpak details."""
    first_app_config.flatpak_runtime = "org.beeware.Platform"
    first_app_config.flatpak_runtime_version = "37.42"
    first_app_config.flatpak_sdk = "org.beeware.SDK"

    assert create_command.output_format_template_context(first_app_config) == {
        "flatpak_runtime": "org.beeware.Platform",
        "flatpak_runtime_version": "37.42",
        "flatpak_sdk": "org.beeware.SDK",
    }


DEFAULT_FINISH_ARGS = {
    "share=ipc": True,
    "socket=x11": True,
    "nosocket=wayland": True,
    "share=network": True,
    "device=dri": True,
    "socket=pulseaudio": True,
    "filesystem=xdg-cache": True,
    "filesystem=xdg-config": True,
    "filesystem=xdg-data": True,
    "filesystem=xdg-documents": True,
    "socket=session-bus": True,
}


@pytest.mark.parametrize(
    "permissions, finish_args, context",
    [
        # No permissions
        (
            {},
            {},
            {
                "finish_args": DEFAULT_FINISH_ARGS,
            },
        ),
        # Only custom permissions
        (
            {
                "custom_permission": "Custom message",
            },
            {
                "allow=bluetooth": True,
            },
            {
                "finish_args": {
                    "share=ipc": True,
                    "socket=x11": True,
                    "nosocket=wayland": True,
                    "share=network": True,
                    "device=dri": True,
                    "socket=pulseaudio": True,
                    "filesystem=xdg-cache": True,
                    "filesystem=xdg-config": True,
                    "filesystem=xdg-data": True,
                    "filesystem=xdg-documents": True,
                    "socket=session-bus": True,
                    "allow=bluetooth": True,
                },
            },
        ),
        # Camera permissions
        (
            {
                "camera": "I need to see you",
            },
            {},
            {
                "finish_args": DEFAULT_FINISH_ARGS,
            },
        ),
        # Microphone permissions
        (
            {
                "microphone": "I need to hear you",
            },
            {},
            {
                "finish_args": DEFAULT_FINISH_ARGS,
            },
        ),
        # Coarse location permissions
        (
            {
                "coarse_location": "I need to know roughly where you are",
            },
            {},
            {
                "finish_args": DEFAULT_FINISH_ARGS,
            },
        ),
        # Fine location permissions
        (
            {
                "fine_location": "I need to know exactly where you are",
            },
            {},
            {
                "finish_args": DEFAULT_FINISH_ARGS,
            },
        ),
        # Background location permissions
        (
            {
                "background_location": "I always need to know where you are",
            },
            {},
            {
                "finish_args": DEFAULT_FINISH_ARGS,
            },
        ),
        # Coarse location background permissions
        (
            {
                "coarse_location": "I need to know roughly where you are",
                "background_location": "I always need to know where you are",
            },
            {},
            {
                "finish_args": DEFAULT_FINISH_ARGS,
            },
        ),
        # Fine location background permissions
        (
            {
                "fine_location": "I need to know exactly where you are",
                "background_location": "I always need to know where you are",
            },
            {},
            {
                "finish_args": DEFAULT_FINISH_ARGS,
            },
        ),
        # Coarse and fine location permissions
        (
            {
                "coarse_location": "I need to know roughly where you are",
                "fine_location": "I need to know exactly where you are",
            },
            {},
            {
                "finish_args": DEFAULT_FINISH_ARGS,
            },
        ),
        # Coarse and fine background location permissions
        (
            {
                "coarse_location": "I need to know roughly where you are",
                "fine_location": "I need to know exactly where you are",
                "background_location": "I always need to know where you are",
            },
            {},
            {
                "finish_args": DEFAULT_FINISH_ARGS,
            },
        ),
        # Photo library permissions
        (
            {
                "photo_library": "I need to see your library",
            },
            {},
            {
                "finish_args": DEFAULT_FINISH_ARGS,
            },
        ),
        # Override and augment by cross-platform definitions
        (
            {
                "fine_location": "I need to know where you are",
                "NSCustomMessage": "Custom message",
            },
            {
                "socket=pulseaudio": False,
                "allow=bluetooth": True,
            },
            {
                "finish_args": {
                    "share=ipc": True,
                    "socket=x11": True,
                    "nosocket=wayland": True,
                    "share=network": True,
                    "device=dri": True,
                    "socket=pulseaudio": False,
                    "filesystem=xdg-cache": True,
                    "filesystem=xdg-config": True,
                    "filesystem=xdg-data": True,
                    "filesystem=xdg-documents": True,
                    "socket=session-bus": True,
                    "allow=bluetooth": True,
                },
            },
        ),
    ],
)
def test_permissions_context(
    create_command, first_app, permissions, finish_args, context
):
    """Platform-specific permissions can be added to the context."""
    # Set the permission and entitlement value
    first_app.permission = permissions
    first_app.finish_arg = finish_args
    # Extract the cross-platform permissions
    x_permissions = create_command._x_permissions(first_app)
    # Check that the final platform permissions are rendered as expected.
    assert context == create_command.permissions_context(first_app, x_permissions)


def test_missing_runtime_config(create_command, first_app_config):
    """The app creation errors is a Flatpak runtime is not defined."""
    create_command.tools.flatpak = MagicMock(spec_set=Flatpak)

    with pytest.raises(
        BriefcaseConfigError,
        match="Briefcase configuration error: The App does not specify the Flatpak runtime to use",
    ):
        create_command.output_format_template_context(first_app_config)



# The package command inherits most of its behavior from the common base
# implementation. Do a surface-level verification here, but the app
# tests provide the actual test coverage.
import os
from pathlib import Path
from unittest import mock

import pytest

from briefcase.console import Console, Log
from briefcase.integrations.subprocess import Subprocess
from briefcase.integrations.wix import WiX
from briefcase.platforms.windows.visualstudio import WindowsVisualStudioPackageCommand


@pytest.fixture
def package_command(tmp_path):
    command = WindowsVisualStudioPackageCommand(
        logger=Log(),
        console=Console(),
        base_path=tmp_path / "base_path",
        data_path=tmp_path / "briefcase",
    )
    command.tools.subprocess = mock.MagicMock(spec_set=Subprocess)
    command.tools.wix = WiX(command.tools, wix_home=tmp_path / "wix")
    return command


def test_package_msi(package_command, first_app_config, tmp_path):
    """A Windows app can be packaged as an MSI."""

    package_command.package_app(first_app_config)

    package_command.tools.subprocess.run.assert_has_calls(
        [
            # Collect manifest
            mock.call(
                [
                    tmp_path / "wix/bin/heat.exe",
                    "dir",
                    Path("x64/Release"),
                    "-nologo",
                    "-gg",
                    "-sfrag",
                    "-sreg",
                    "-srd",
                    "-scom",
                    "-dr",
                    "first_app_ROOTDIR",
                    "-cg",
                    "first_app_COMPONENTS",
                    "-var",
                    "var.SourceDir",
                    "-out",
                    "first-app-manifest.wxs",
                ],
                check=True,
                cwd=tmp_path
                / "base_path"
                / "build"
                / "first-app"
                / "windows"
                / "visualstudio",
            ),
            # Compile MSI
            mock.call(
                [
                    tmp_path / "wix/bin/candle.exe",
                    "-nologo",
                    "-ext",
                    "WixUtilExtension",
                    "-ext",
                    "WixUIExtension",
                    "-arch",
                    "x64",
                    f'-dSourceDir={os.fsdecode(Path("x64/Release"))}',
                    "first-app.wxs",
                    "first-app-manifest.wxs",
                ],
                check=True,
                cwd=tmp_path
                / "base_path"
                / "build"
                / "first-app"
                / "windows"
                / "visualstudio",
            ),
            # Link MSI
            mock.call(
                [
                    tmp_path / "wix/bin/light.exe",
                    "-nologo",
                    "-ext",
                    "WixUtilExtension",
                    "-ext",
                    "WixUIExtension",
                    "-loc",
                    "unicode.wxl",
                    "-o",
                    tmp_path / "base_path/dist/First App-0.0.1.msi",
                    "first-app.wixobj",
                    "first-app-manifest.wixobj",
                ],
                check=True,
                cwd=tmp_path
                / "base_path"
                / "build"
                / "first-app"
                / "windows"
                / "visualstudio",
            ),
        ]
    )



import os
import subprocess
from pathlib import Path
from unittest.mock import ANY

import pytest

from briefcase.console import LogLevel


def test_call(mock_sub, sub_stream_kw, sleep_zero, capsys):
    """A simple call will be invoked."""

    with mock_sub.tools.input.wait_bar():
        mock_sub.run(["hello", "world"])

    mock_sub._subprocess.Popen.assert_called_with(["hello", "world"], **sub_stream_kw)
    # fmt: off
    expected_output = (
        "output line 1\n"
        "\n"
        "output line 3\n"
        "\n"
    )
    # fmt: on
    assert capsys.readouterr().out == expected_output


def test_call_with_arg(mock_sub, sub_stream_kw, sleep_zero, capsys):
    """Any extra keyword arguments are passed through as-is."""

    with mock_sub.tools.input.wait_bar():
        mock_sub.run(["hello", "world"], extra_kw="extra")

    mock_sub._subprocess.Popen.assert_called_with(
        ["hello", "world"],
        extra_kw="extra",
        **sub_stream_kw,
    )
    # fmt: off
    expected_output = (
        "output line 1\n"
        "\n"
        "output line 3\n"
        "\n"
    )
    # fmt: on
    assert capsys.readouterr().out == expected_output


def test_debug_call(mock_sub, sub_stream_kw, sleep_zero, capsys):
    """If verbosity is turned up, there is debug output."""
    mock_sub.tools.logger.verbosity = LogLevel.DEBUG

    with mock_sub.tools.input.wait_bar():
        mock_sub.run(["hello", "world"])

    mock_sub._subprocess.Popen.assert_called_with(["hello", "world"], **sub_stream_kw)
    expected_output = (
        "\n"
        ">>> Running Command:\n"
        ">>>     hello world\n"
        ">>> Working Directory:\n"
        f">>>     {Path.cwd()}\n"
        "output line 1\n"
        "\n"
        "output line 3\n"
        ">>> Return code: -3\n"
        "\n"
    )
    assert capsys.readouterr().out == expected_output


def test_debug_call_with_env(mock_sub, sub_stream_kw, sleep_zero, capsys, tmp_path):
    """If verbosity is turned up, injected env vars are included in debug output."""
    mock_sub.tools.logger.verbosity = LogLevel.DEBUG

    env = {"NewVar": "NewVarValue"}
    with mock_sub.tools.input.wait_bar():
        mock_sub.run(
            ["hello", "world"],
            env=env,
            cwd=tmp_path / "cwd",
        )

    merged_env = mock_sub.tools.os.environ.copy()
    merged_env.update(env)

    mock_sub._subprocess.Popen.assert_called_with(
        ["hello", "world"],
        env=merged_env,
        cwd=os.fsdecode(tmp_path / "cwd"),
        **sub_stream_kw,
    )
    expected_output = (
        "\n"
        ">>> Running Command:\n"
        ">>>     hello world\n"
        ">>> Working Directory:\n"
        f">>>     {tmp_path / 'cwd'}\n"
        ">>> Environment Overrides:\n"
        ">>>     NewVar=NewVarValue\n"
        "output line 1\n"
        "\n"
        "output line 3\n"
        ">>> Return code: -3\n"
        "\n"
    )
    assert capsys.readouterr().out == expected_output


@pytest.mark.parametrize(
    "in_kwargs, kwargs",
    [
        (
            {},
            {"text": True, "encoding": ANY, "bufsize": 1, "errors": "backslashreplace"},
        ),
        (
            {"text": True},
            {"text": True, "encoding": ANY, "bufsize": 1, "errors": "backslashreplace"},
        ),
        ({"text": False}, {"text": False, "bufsize": 1}),
        ({"text": False, "bufsize": 42}, {"text": False, "bufsize": 42}),
        ({"universal_newlines": False}, {"text": False, "bufsize": 1}),
        (
            {"universal_newlines": True},
            {"text": True, "encoding": ANY, "bufsize": 1, "errors": "backslashreplace"},
        ),
    ],
)
def test_text_eq_true_default_overriding(mock_sub, in_kwargs, kwargs, sleep_zero):
    """If text or universal_newlines is explicitly provided, those should override
    text=true default and universal_newlines should be converted to text."""
    with mock_sub.tools.input.wait_bar():
        mock_sub.run(["hello", "world"], **in_kwargs)

    mock_sub._subprocess.Popen.assert_called_with(
        ["hello", "world"],
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        **kwargs,
    )


def test_stderr_is_redirected(
    mock_sub,
    streaming_process,
    sub_stream_kw,
    sleep_zero,
    capsys,
):
    """When stderr is redirected, it should be included in the result."""
    stderr_output = "stderr output\nline 2"
    streaming_process.stderr.read.return_value = stderr_output

    with mock_sub.tools.input.wait_bar():
        run_result = mock_sub.run(
            ["hello", "world"],
            stderr=subprocess.PIPE,
        )

    sub_stream_kw.pop("stderr")
    mock_sub._subprocess.Popen.assert_called_with(
        ["hello", "world"],
        stderr=subprocess.PIPE,
        **sub_stream_kw,
    )
    # fmt: off
    expected_output = (
        "output line 1\n"
        "\n"
        "output line 3\n"
        "\n"
    )
    # fmt: on
    assert capsys.readouterr().out == expected_output
    assert run_result.stderr == stderr_output


def test_stderr_dev_null(
    mock_sub,
    streaming_process,
    sub_stream_kw,
    sleep_zero,
    capsys,
):
    """When stderr is discarded, it should be None in the result."""
    streaming_process.stderr = None

    with mock_sub.tools.input.wait_bar():
        run_result = mock_sub.run(
            ["hello", "world"],
            stderr=subprocess.DEVNULL,
        )

    sub_stream_kw.pop("stderr")
    mock_sub._subprocess.Popen.assert_called_with(
        ["hello", "world"],
        stderr=subprocess.DEVNULL,
        **sub_stream_kw,
    )
    # fmt: off
    expected_output = (
        "output line 1\n"
        "\n"
        "output line 3\n"
        "\n"
    )
    # fmt: on
    assert capsys.readouterr().out == expected_output
    assert run_result.stderr is None


def test_calledprocesserror(mock_sub, streaming_process, sleep_zero, capsys):
    """CalledProcessError is raised with check=True and non-zero return value."""
    stderr_output = "stderr output\nline 2"
    streaming_process.stderr.read.return_value = stderr_output

    with pytest.raises(subprocess.CalledProcessError) as exc:
        with mock_sub.tools.input.wait_bar():
            mock_sub.run(
                ["hello", "world"],
                check=True,
                stderr=subprocess.PIPE,
            )

    # fmt: off
    expected_output = (
        "output line 1\n"
        "\n"
        "output line 3\n"
        "\n"
    )
    # fmt: on
    assert capsys.readouterr().out == expected_output
    assert exc.value.returncode == -3
    assert exc.value.cmd == ["hello", "world"]
    assert exc.value.stderr == stderr_output


def test_invalid_invocations(mock_sub):
    """Ensure run cannot be used in a Wait Bar with invalid arguments."""
    with pytest.raises(AssertionError):
        with mock_sub.tools.input.wait_bar():
            mock_sub.run(
                ["hello", "world"],
                stdout=subprocess.PIPE,
            )

    for invalid_arg in ("timeout", "input"):
        with pytest.raises(AssertionError):
            with mock_sub.tools.input.wait_bar():
                mock_sub.run(["hello", "world"], **{invalid_arg: "value"})



import sys
from pathlib import Path
from unittest.mock import MagicMock

import pytest

from briefcase.exceptions import BriefcaseCommandError, NetworkFailure

from ....utils import create_tgz_file


@pytest.fixture
def skin_tgz_path(tmp_path) -> Path:
    return create_tgz_file(tmp_path / "skin.tgz", content=[("skin", "skin contents")])


def test_existing_skin(mock_tools, android_sdk):
    """If the skin already exists, don't attempt to download it again."""
    # Mock the existence of a system image
    (android_sdk.root_path / "skins/pixel_X").mkdir(parents=True)

    # Verify the system image that we already have
    android_sdk.verify_emulator_skin("pixel_X")

    # file.download was *not* called.
    mock_tools.file.download.assert_not_called()


def test_new_skin(mock_tools, android_sdk, skin_tgz_path):
    """If the skin doesn't exist, an attempt is made to download it."""
    mock_tools.file.download = MagicMock(return_value=skin_tgz_path)

    # Verify the skin, triggering a download
    android_sdk.verify_emulator_skin("pixel_X")

    # Skin was downloaded
    mock_tools.file.download.assert_called_once_with(
        url="https://android.googlesource.com/platform/tools/adt/idea/"
        "+archive/refs/heads/mirror-goog-studio-main/"
        "artwork/resources/device-art-resources/pixel_X.tar.gz",
        download_path=android_sdk.root_path,
        role="pixel_X device skin",
    )

    # Skin is unpacked.
    mock_tools.shutil.unpack_archive.assert_called_once_with(
        filename=skin_tgz_path,
        extract_dir=android_sdk.root_path / "skins/pixel_X",
        **({"filter": "data"} if sys.version_info >= (3, 12) else {}),
    )

    # Original file was deleted.
    assert not skin_tgz_path.exists()


def test_skin_download_failure(mock_tools, android_sdk, skin_tgz_path):
    """If the skin download fails, an error is raised."""
    mock_tools.file.download = MagicMock(return_value=skin_tgz_path)

    # Mock a failure downloading the skin
    mock_tools.file.download.side_effect = NetworkFailure("mock")

    # Verify the skin, triggering a download
    with pytest.raises(NetworkFailure, match="Unable to mock"):
        android_sdk.verify_emulator_skin("pixel_X")

    # An attempt was made to download the skin
    mock_tools.file.download.assert_called_once_with(
        url="https://android.googlesource.com/platform/tools/adt/idea/"
        "+archive/refs/heads/mirror-goog-studio-main/"
        "artwork/resources/device-art-resources/pixel_X.tar.gz",
        download_path=android_sdk.root_path,
        role="pixel_X device skin",
    )

    # Skin wasn't downloaded, so it wasn't unpacked
    assert mock_tools.shutil.unpack_archive.call_count == 0


def test_unpack_failure(mock_tools, android_sdk, skin_tgz_path):
    """If the download is corrupted and unpacking fails, an error is raised."""
    # Mock the result of the download of a skin
    mock_tools.file.download = MagicMock(return_value=skin_tgz_path)

    # Mock a failure unpacking the skin
    mock_tools.shutil.unpack_archive.side_effect = EOFError

    # Verify the skin, triggering a download
    with pytest.raises(
        BriefcaseCommandError,
        match=r"Unable to unpack pixel_X device skin.",
    ):
        android_sdk.verify_emulator_skin("pixel_X")

    # Skin was downloaded
    mock_tools.file.download.assert_called_once_with(
        url="https://android.googlesource.com/platform/tools/adt/idea/"
        "+archive/refs/heads/mirror-goog-studio-main/"
        "artwork/resources/device-art-resources/pixel_X.tar.gz",
        download_path=android_sdk.root_path,
        role="pixel_X device skin",
    )

    # An attempt to unpack the skin was made.
    mock_tools.shutil.unpack_archive.assert_called_once_with(
        filename=skin_tgz_path,
        extract_dir=android_sdk.root_path / "skins/pixel_X",
        **({"filter": "data"} if sys.version_info >= (3, 12) else {}),
    )

    # Original file wasn't deleted.
    assert skin_tgz_path.exists()



import pytest

from ....utils import create_file, create_wheel


@pytest.fixture
def first_app_generated(first_app_config, tmp_path):
    # Create the briefcase.toml file
    bundle_path = tmp_path / "base_path/build/first-app/web/static"
    create_file(
        bundle_path / "briefcase.toml",
        """
[paths]
app_path="app"
app_requirements_path="requirements.txt"
""",
    )

    # Create index.html
    create_file(bundle_path / "www/index.html", "<html></html>")

    # Create the initial briefcase.css
    create_file(
        bundle_path / "www/static/css/briefcase.css",
        """
#pyconsole {
  display: None;
}
/*******************************************************************
 ******************** Wheel contributed styles ********************/
""",
    )

    # Create an empty wheels folder
    (bundle_path / "www/static/wheels").mkdir(parents=True)

    return first_app_config


@pytest.fixture
def first_app_built(first_app_generated, tmp_path):
    bundle_path = tmp_path / "base_path/build/first-app/web/static"

    # Create pyscript.toml
    create_file(
        bundle_path / "www/pyscript.toml",
        'packages = ["dummy-1.2.3-py3-none-all.whl"]',
    )

    # Create an app wheel
    create_wheel(bundle_path / "www/static/wheels")

    return first_app_generated



from __future__ import annotations

import os
import shutil
import sys
import tempfile
from contextlib import suppress
from email.message import Message
from pathlib import Path
from urllib.parse import urlparse

import requests.exceptions as requests_exceptions
from requests import Response

from briefcase.exceptions import (
    BadNetworkResourceError,
    MissingNetworkResourceError,
    NetworkFailure,
)
from briefcase.integrations.base import Tool, ToolCache


class File(Tool):
    name = "file"
    full_name = "File"

    @classmethod
    def verify_install(cls, tools: ToolCache, **kwargs) -> File:
        """Make File available in tool cache."""
        # short circuit since already verified and available
        if hasattr(tools, "file"):
            return tools.file

        tools.file = File(tools=tools)
        return tools.file

    def is_archive(self, filename: str | os.PathLike) -> bool:
        """Can a file be unpacked via `shutil.unpack_archive()`?

        The evaluation is based purely on the name of the file. A more robust
        implementation may actually interrogate the file itself.

        Notably, the types of archives that can be unpacked will vary based on the types
        of compression the current Python supports. So, for example, if LZMA is not
        available, then tar.xz archives cannot be unpacked.

        :param filename: path to file to evaluate as an archive
        """
        filename = Path(filename)
        file_extensions = {
            # captures extensions like .tar.gz, .tar.bz2, etc.
            "".join(filename.suffixes[-2:]),
            # as well as .tar, .zip, etc.
            filename.suffix,
        }
        return not file_extensions.isdisjoint(self.supported_archive_extensions)

    @property
    def supported_archive_extensions(self) -> set[str]:
        return {
            extension
            for archive_format in shutil.get_unpack_formats()
            for extension in archive_format[1]
        }

    def unpack_archive(
        self,
        filename: str | os.PathLike,
        extract_dir: str | os.PathLike,
        **kwargs,
    ):
        """Unpack an archive file in to a destination directory.

        Additional protections for unpacking tar files were introduced in Python 3.12.
        Since tarballs can contain anything valid in a UNIX file system, these
        protections prevent unpacking potentially dangerous files. This behavior will be
        the default in Python 3.14. However, the protections can only be enabled for tar
        files...not zip files.

        :param filename: File path for the archive
        :param extract_dir: Target file path for where to unpack archive
        :param kwargs: additional arguments for shutil.unpack_archive
        """
        is_zip = Path(filename).suffix == ".zip"
        if sys.version_info >= (3, 12):  # pragma: no-cover-if-lt-py312
            unpack_kwargs = {"filter": "data"} if not is_zip else {}
        else:  # pragma: no-cover-if-gte-py312
            unpack_kwargs = {}

        self.tools.shutil.unpack_archive(
            filename=filename,
            extract_dir=extract_dir,
            **{
                **unpack_kwargs,
                **kwargs,
            },
        )

    def download(self, url: str, download_path: Path, role: str | None = None) -> Path:
        """Download a given URL, caching it. If it has already been downloaded, return
        the value that has been cached.

        This is a utility method used to obtain assets used by the installation process.
        The cached filename will be the filename portion of the URL, appended to the
        download path.

        :param url: The URL to download
        :param download_path: The path to the download cache folder. This path will be
            created if it doesn't exist.
        :param role: A string describing the role played by the file being downloaded;
            used to construct log and error messages. Should be able to fit into the
            sentence "Error downloading {role}".
        :returns: The filename of the downloaded (or cached) file.
        """
        download_path.mkdir(parents=True, exist_ok=True)
        filename: Path = None
        try:
            response = self.tools.requests.get(url, stream=True)
            if response.status_code == 404:
                raise MissingNetworkResourceError(url=url)
            elif response.status_code != 200:
                raise BadNetworkResourceError(url=url, status_code=response.status_code)

            # The initial URL might (read: will) go through URL redirects, so
            # we need the *final* response. We look at either the `Content-Disposition`
            # header, or the final URL, to extract the cache filename.
            cache_full_name = urlparse(response.url).path
            header_value = response.headers.get("Content-Disposition")
            if header_value:
                # Neither requests nor httplib provides a way to parse RFC6266 headers.
                # The cgi module *did* have a way to parse these headers, but
                # it was deprecated as part of PEP594. PEP594 recommends
                # using the email.message module to parse these headers as they
                # are near identical format.
                # See also:
                # * https://tools.ietf.org/html/rfc6266
                # * https://peps.python.org/pep-0594/#cgi
                msg = Message()
                msg["Content-Disposition"] = header_value
                filename = msg.get_filename()
                if filename:
                    cache_full_name = filename
            cache_name = cache_full_name.split("/")[-1]
            filename = download_path / cache_name

            if filename.exists():
                self.tools.logger.info(f"{cache_name} already downloaded")
            else:
                self.tools.logger.info(f"Downloading {cache_name}...")
                self._fetch_and_write_content(response, filename)
        except requests_exceptions.ConnectionError as e:
            if role:
                description = role
            else:
                description = filename.name if filename else url
            raise NetworkFailure(f"download {description}") from e

        return filename

    def _fetch_and_write_content(self, response: Response, filename: Path):
        """Write the content from the requests Response to file.

        The data is initially written in to a temporary file in the Briefcase
        cache. This avoids partially downloaded files masquerading as complete
        downloads in later Briefcase runs. The temporary file is only moved
        to ``filename`` if the download is successful; otherwise, it is deleted.

        :param response: ``requests.Response``
        :param filename: full filesystem path to save data
        """
        temp_file = tempfile.NamedTemporaryFile(
            dir=filename.parent,
            prefix=f"{filename.name}.",
            suffix=".download",
            delete=False,
        )
        try:
            with temp_file:
                total = response.headers.get("content-length")
                if total is None:
                    temp_file.write(response.content)
                else:
                    progress_bar = self.tools.input.progress_bar()
                    task_id = progress_bar.add_task("Downloader", total=int(total))
                    with progress_bar:
                        for data in response.iter_content(chunk_size=1024 * 1024):
                            temp_file.write(data)
                            progress_bar.update(task_id, advance=len(data))

            # This file move short circuits to a file rename when the source and
            # destination are on the same filesystem; therefore, it should complete
            # quite quickly even for large files.
            self.tools.shutil.move(temp_file.name, filename)
            # Temporary files are created with only read/write permissions for the
            # file's owner (i.e. 600); to match the behavior of file creation using
            # ``open(..., "w")``, the downloaded file's permissions are updated for
            # the group and world to have read/write permissions as well. Finally,
            # (as with ``open()``) the system's umask is respected. The current
            # umask is only available as the return value to updating the umask...
            # Updating the umask affects the current process, including all threads.
            # A umask value represents permissions that should be denied; so, 022
            # denies write permissions to the group and world. A 022 umask is a
            # common default among supporting systems.
            os.umask(current_umask := os.umask(0o022))
            # The umask is applied by inverting it and bitwise ANDing the default
            # permissions...thus masking out permissions that should be denied.
            self.tools.os.chmod(filename, 0o666 & ~current_umask)
        finally:
            # Ensure the temporary file is deleted; this file may still
            # exist if the download fails or the user sends CTRL+C.
            with suppress(FileNotFoundError):
                self.tools.os.remove(temp_file.name)



import subprocess
from datetime import datetime
from unittest.mock import MagicMock, PropertyMock

import pytest

from briefcase.exceptions import BriefcaseCommandError


@pytest.mark.parametrize("is_color_enabled", [True, False])
def test_logcat_tail(mock_tools, adb, is_color_enabled, monkeypatch):
    """Invoking `logcat_tail()` calls `run()` with the appropriate parameters."""
    # Mock whether color is enabled for the console
    monkeypatch.setattr(
        type(mock_tools.input),
        "is_color_enabled",
        PropertyMock(return_value=is_color_enabled),
    )

    # Invoke logcat_tail with a specific timestamp
    adb.logcat_tail(since=datetime(2022, 11, 10, 9, 8, 7))

    # Validate call parameters.
    mock_tools.subprocess.run.assert_called_once_with(
        [
            mock_tools.android_sdk.adb_path,
            "-s",
            "exampleDevice",
            "logcat",
            "--format=tag",
            "-t",
            "11-10 09:08:07.000000",
            "-s",
            "MainActivity:*",
            "stdio:*",
            "python.stdout:*",
            "AndroidRuntime:*",
        ]
        + (["--format=color"] if is_color_enabled else []),
        env=mock_tools.android_sdk.env,
        check=True,
        encoding="UTF-8",
    )


def test_adb_failure(mock_tools, adb):
    """If adb logcat fails, the error is caught."""
    # Mock out the run command on an adb instance
    mock_tools.subprocess.run = MagicMock(
        side_effect=subprocess.CalledProcessError(returncode=1, cmd="adb logcat")
    )

    with pytest.raises(BriefcaseCommandError):
        adb.logcat_tail(since=datetime(2022, 11, 10, 9, 8, 7))
